# coding: utf-8
#
# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# json schema represented as a straight-forward python datasture, e.g.:
#
# {
#   'name': str,
#   'age': int
# }
#
# ... means the corresponding json object must be a dictionary with
# "name" and "age" items with the specified types (string and int respectively).
#
from xju.xn import Xn,in_context,in_function_context,first_line_of,readable_repr

l1=first_line_of

class Schema:
    def __init__(self,x):
        'initialse schema %(x)r'
        validateSchemaElement(x)
        self.x=x
        pass
    def __repr__(self):
        return repr(self.x)
    def validate(self,x):
        'verify that %(x)r conforms to jsonschema.Schema %(self)r'%vars()
        return validate(self.x,x)
    pass

class OneOf:
    def __init__(self,*choices):
        self.choices=choices
        pass
    def __str__(self):
        return 'one of %(choices)s'%self.__dict__
    def __repr__(self):
        return 'one of %(choices)r'%self.__dict__
    pass

def validateSchemaElement(x):
    'verify that {x!r} is a valid json schema element'
    try:
        if x is None: return
        if x in [int,str,float,bool]: return
        if type(x) in [int,str]: return #literal
        if type(x) is dict:
            if len(x) == 1 and list(x.keys())[0] in (int,str):
                validateSchemaElement(list(x.values())[0])
                return
            for name,y in x.items():
                try:
                    if type(name) is not str:
                        raise Exception(f'{name} is not a string (it is a {name.__class__.__name__})')
                    validateSchemaElement(y)
                except:
                    raise in_context('validate dict schema item %(name)r'%vars()) from None
                pass
            return
        if type(x) is list:
            if len(x) == 0:
                raise Exception(
                    'list schema must contain at least one element')
            for xx in x: validateSchemaElement(xx)
            return
        if type(x) is tuple:
            for i,y in enumerate(x):
                try:
                    validateSchemaElement(y)
                except:
                    raise in_context('validate tuple schema item %(i)r'%vars()) from None
                pass
            return
        if isinstance(x,OneOf):
            for c in x.choices:
                validateSchemaElement(c)
                pass
            return
        if type(x) is bool:
            #fixed literal value True or False
            return
        if isinstance(x,Schema):
            return
        t=type(x)
        if t is object: t=x.__class__
        raise Exception(f'jsonschema element may not be a {t}, it must be a list, a dictionary or int, str, float, bool, tuple, {__name__}.OneOf or None')
    except:
        raise in_function_context(validateSchemaElement,vars()) from None
    pass

def validate(schema,x):
    'verify {x!r} conforms to json schema {schema!r}'
    try:
        if schema is None and not x is None:
            raise Exception('%(x)r is not None'%vars())
        if type(schema) is int:
            if not type(x) is int:
                raise Exception('%(x)r is not an integer'%vars())
            if not x==schema:
                raise Exception('%(x)r is not %(schema)r'%vars())
            pass
        if type(schema) in [str]:
            if not type(x) in [str]:
                raise Exception('%(x)r is not a string'%vars())
            if not x==schema:
                raise Exception('%(x)r is not %(schema)r'%vars())
            pass
        if schema is int and not type(x) is int:
            raise Exception('%(x)r is not an Int'%vars())
        if schema is bool and not type(x) is bool:
            raise Exception('%(x)r is not a Boolean'%vars())
        if schema is float and not type(x) in [int,float]:
            raise Exception('%(x)r is not a Float'%vars())
        if schema is str and not type(x) in [str]:
            raise Exception('%(x)r is not a String'%vars())
        if type(schema) is dict:
            if not type(x) is dict:
                raise Exception('%(x)r is not a Dictionary'%vars())
            if len(schema)==1 and list(schema.keys())[0] in (int,str):
                for key,y in x.items():
                    try:
                        validate(list(schema.keys())[0],key)
                        validate(list(schema.values())[0],y)
                    except:
                        raise in_context('validate dictionary item %(key)r'%vars()) from None
                    pass
                return x
            for name, y in schema.items():
                try:
                    if not name in x:
                        try:
                            validate(y,None)
                        except:
                            keys=list(x.keys())
                            raise Exception('%(name)r is not in %(keys)r and %(name)r is not optional'%vars())
                        pass
                    else:
                        validate(y,x[name])
                        pass
                except:
                    raise in_context('validate dictionary item %(name)r'%vars()) from None
                pass
            return x
        if type(schema) is list:
            if not type(x) is list:
                raise Exception('%(x)r is not a List'%vars())
            if len(schema)==1:
                for i,y in enumerate(x):
                    try:
                        validate(schema[0],y)
                    except:
                        raise in_context('validate list item %(i)r'%vars()) from None
                    pass
                pass
            else:
                if len(schema)!=len(x):
                    i=len(schema)
                    j=len(x)
                    raise Exception('list {x} does not have {i} elements, it has {j} elements'.format(**vars()))
                for i,v in enumerate(zip(schema,x)):
                    try:
                        validate(v[0],v[1])
                    except:
                        raise in_context('validate list element {i}'.format(**vars())) from None
                pass
            pass
        if type(schema) is tuple:
            if len(schema) != len(x):
                sl=len(schema)
                xl=len(x)
                raise Exception('tuple has %(xl)s items not %(sl)s'%vars())
            for i,y in enumerate(x):
                try:
                    validate(schema[i],y)
                except:
                    raise in_context('validate tuple schema element %(i)s'%vars()) from None
                pass
            pass
        if isinstance(schema,OneOf):
            choices=schema.choices[:]
            failures=[]
            while len(choices):
                try:
                    validate(choices[0],x)
                    return x
                except Exception as e:
                    failures.append(e)
                    pass
                choices=choices[1:]
                pass
            raise Exception(' and '.join([readable_repr(_) for _ in failures]))
        if type(schema) is bool:
            if x==schema: return x
            raise Exception('{x!r} is not {schema!r}'.format(**vars()))
        if isinstance(schema,Schema):
            schema.validate(x)
            pass
        return x
    except:
        raise in_function_context(validate,vars()) from None
    pass
                    




# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# Utilities for creating distinct types mirroring a subset of a basic
# type (str, int, float).
#
# Unlike typing.NewType:
#   - works with isinstance (at runtime)
#   - many methods of the basic type are provided directly, e.g. for
#     new int type A, A+A->A
#
# For example to define an int-like Hours class:
#
#  class HoursTag:pass
#  class Hours(Int[HoursTag]):pass
#
# ... note do not use 'Hours=Int[HoursTag]' because that is an alias to a generic and
# therefore has not run-time presence and therefore cannot be used with isinstance.
#
from typing import Iterable,Sized,Container,Collection,Reversible,Protocol,Type,overload,TypeVar
from typing import Generic,Tuple,Mapping,Optional,List,Literal,Union,Any,Self,Never
from types import NotImplementedType

Tag=TypeVar('Tag',covariant=True)

class Int(Generic[Tag]):
    __value:int

    def __init__(self, value:int):
        self.__value=value
        pass

    def value(self)->int:
        return self.__value

    # note the following type: ignore is to get the desired behaviour from mypy --strict-equality
    # (as at mypy 1.3.0) i.e. forbid comparison of Int[XTag] and Int[YTag], noting:
    # - python itself insists on being able to compare values of unrelated type (for example
    #   to implement x in y)
    # - the type: ignore avoid mypy error for non-liskoff substitutability (because
    #   python's Object.__eq__ has other as Any)
    # - with other: Any, mypy --strict-equality gives no error at all
    # - magically (incorrectly?), the type: ignore does not suppress checks on type of other, but
    #   does suppress the substitution error
    def __eq__(self,other:Self)->bool:  # type: ignore
        '''equality test, self and other have the same type and same value'''
        '''i.e. recommend stick to using Int[X] like:
              class Hours(Int[HoursTag]):pass
           ... and not inherit from Hours.
           If you choose to inherit from Hours, make sure you write your own __eq__'''
        if type(other) is not type(self):
            return False
        return self.value().__eq__(other.value())

    def __ne__(self,other:Self)->bool:  # type: ignore
        return not self.__eq__(other)

    def __str__(self)->str:
        return str(self.value())

    def __repr__(self)->str:
        return repr(self.value())

    def __format__(self, format_spec:str)->str:
        return self.value().__format__(format_spec)

    def __float__(self)->float:
        return self.value().__float__()
    
    @overload
    def __divmod__(self, x:int) -> Tuple[Self,Self]:
        pass
    @overload
    def __divmod__(self, x:float) -> Tuple[float,float]:
        pass
    @overload
    def __divmod__(self, x:Self) -> Tuple[int,int]:
        pass
    def __divmod__(self, x:Any) -> Any:
        if isinstance(x,int):
            q,r=self.value().__divmod__(x)
            return self.__class__(q),self.__class__(r)
        if isinstance(x,float):
            return divmod(self.value(),x)
        else:
            return divmod(self.value(),x.value())
        pass

    @overload
    def __floordiv__(self, x:int) -> Self:
        pass
    @overload
    def __floordiv__(self, x:float) -> float:
        pass
    @overload
    def __floordiv__(self, x:Self) -> int:
        pass
    def __floordiv__(self, x:Any) -> Any:
        if isinstance(x,int):
            return self.__class__(self.value()//x)
        elif isinstance(x,float):
            return self.value()//x
        else:
            return self.value()//x.value()
        pass

    def __truediv__(self, x:float|int|Self) -> float:
        if isinstance(x,float) or isinstance(x,int):
            return self.value()/x
        else:
            return self.value()/x.value()
        pass
    
    @overload
    def __mul__(self, x:int) -> Self:
        pass
    @overload
    def __mul__(self, x:float) -> NotImplementedType:
        pass
    def __mul__(self, x:Any) -> Any:
        if isinstance(x,int):
            return self.__class__(self.value()*x)
        return NotImplemented

    @overload
    def __rmul__(self, x:int) -> Self:
        pass
    @overload
    def __rmul__(self, x:float)-> float:
        pass
    def __rmul__(self, x:Any)->Any:
        if isinstance(x,int):
            return self.__class__(x*self.value())
        else:
            return NotImplemented
        pass

    @overload
    def __mod__(self, other:int)->Self:
        pass
    @overload
    def __mod__(self, other:float)->float:
        pass
    @overload
    def __mod__(self, other:Self)->int:
        pass
    def __mod__(self, other:Any)->Any:
        if type(other) is int:
            return self.__class__(self.value()%other)
        if type(other) is float:
            return self.value()%other
        else:
            return self.value()%other.value()

    def __round__(self, ndigits:int=0)->Self:
        return self.__class__(self.value().__round__(ndigits))

    def is_integer(self)->bool:
        return True

    
    def __abs__(self) -> Self:
        return self.__class__(self.value().__abs__())
    def __invert__(self) -> Self:
        return self.__class__(self.value().__invert__())
    def __neg__(self) -> Self:
        return self.__class__(self.value().__neg__())
    def __pos__(self) -> Self:
        return self.__class__(self.value().__pos__())
    def __int__(self)->int:
        return self.value().__int__()
    def __sizeof__(self)->int:
        return self.value().__sizeof__()
    def bit_count(self)->int:
        return self.value().bit_count()
    def bit_length(self)->int:
        return self.value().bit_length()
    def __index__(self)->int:
        return self.value().__index__()
    def __hash__(self)->int:
        return self.value().__hash__()
    def __bool__(self)->bool:
        return self.value().__bool__()
    def __ror__(self,n:int) -> Self:
        return self.__class__(self.value().__ror__(n))
    def __rrshift__(self,n:int) -> Self:
        return self.__class__(self.value().__rrshift__(n))
    def __lshift__(self,n:int) -> Self:
        return self.__class__(self.value().__lshift__(n))
    def __rlshift__(self,n:int) -> Self:
        return self.__class__(self.value().__rlshift__(n))
    def __rshift__(self,n:int) -> Self:
        return self.__class__(self.value().__rshift__(n))
    def __gt__(self,other:Self)->bool:
        return self.value().__gt__(other.value())
    def __lt__(self,other:Self)->bool:
        return self.value().__lt__(other.value())
    def __le__(self,other:Self)->bool:
        return self.value().__le__(other.value())
    def __ge__(self,other:Self)->bool:
        return self.value().__ge__(other.value())
    def __add__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__add__(other.value()))
    def __sub__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__sub__(other.value()))
    def __and__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__and__(other.value()))
    def __or__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__or__(other.value()))
    def __xor__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__xor__(other.value()))
    def as_integer_ratio(self)->Tuple[int,int]:
        return self.value().as_integer_ratio()

    pass


class Float(Generic[Tag]):
    __value:float

    def __init__(self, value:float):
        self.__value=value
        pass

    def value(self)->float:
        return self.__value

    # note the following type: ignore is to get the desired behaviour from mypy --strict-equality
    # (as at mypy 1.3.0) i.e. forbid comparison of Int[XTag] and Int[YTag], noting:
    # - python itself insists on being able to compare values of unrelated type (for example
    #   to implement x in y)
    # - the type: ignore avoid mypy error for non-liskoff substitutability (because
    #   python's Object.__eq__ has other as Any)
    # - with other: Any, mypy --strict-equality gives no error at all
    # - magically (incorrectly?), the type: ignore does not suppress checks on type of other, but
    #   does suppress the substitution error
    def __eq__(self,other:Self)->bool:  # type: ignore
        '''equality test, self and other have the same type and same value'''
        '''i.e. recommend stick to using Int[X] like:
              class Hours(Int[HoursTag]):pass
           ... and not inherit from Hours.
           If you choose to inherit from Hours, make sure you write your own __eq__'''
        if type(other) is not type(self):
            return False
        return self.value().__eq__(other.value())

    def __ne__(self,other:Self)->bool:  # type: ignore
        return not self.__eq__(other)

    def __str__(self)->str:
        return str(self.value())

    def __repr__(self)->str:
        return repr(self.value())

    def __format__(self, format_spec:str)->str:
        return self.value().__format__(format_spec)

    def __int__(self)->int:
        return self.value().__int__()
    
    def __float__(self)->float:
        return self.value().__float__()
    
    def hex(self)->str:
        return self.value().hex()
    
    @overload
    def __divmod__(self, x:int) -> Tuple[Self,Self]:
        pass
    @overload
    def __divmod__(self, x:float) -> Tuple[Self,Self]:
        pass
    @overload
    def __divmod__(self, x:Self) -> Tuple[float,float]:
        pass
    def __divmod__(self, x:Any) -> Any:
        if isinstance(x,int) or isinstance(x,float):
            q,r=self.value().__divmod__(x)
            return self.__class__(q),self.__class__(r)
        else:
            return divmod(self.value(),x.value())
        pass

    @overload
    def __floordiv__(self, x:int) -> Self:
        pass
    @overload
    def __floordiv__(self, x:float) -> Self:
        pass
    @overload
    def __floordiv__(self, x:Self) -> float:
        pass
    def __floordiv__(self, x:Any) -> Any:
        if isinstance(x,int) or isinstance(x,float):
            return self.__class__(self.value()//x)
        else:
            return self.value()//x.value()
        pass

    @overload
    def __truediv__(self, x:Union[float,int]) -> Self:
        pass
    @overload
    def __truediv__(self, x:Self) -> float:
        pass
    def __truediv__(self, x:Any) -> Any:
        if isinstance(x,int) or isinstance(x,float):
            return self.__class__(self.value()/x)
        else:
            return self.value()/x.value()
        pass
    
    @overload
    def __mul__(self, x:int) -> Self:
        pass
    @overload
    def __mul__(self, x:float) -> Self:
        pass
    def __mul__(self, x:Any) -> Any:
        return self.__class__(self.value()*x)

    @overload
    def __rmul__(self, x:int) -> Self:
        pass
    @overload
    def __rmul__(self, x:float) -> Self:
        pass
    def __rmul__(self, x:Any)->Any:
        return self.__class__(x*self.value())

    @overload
    def __mod__(self, other:int)->Self:
        pass
    @overload
    def __mod__(self, other:float)->Self:
        pass
    @overload
    def __mod__(self, other:Self)->float:
        pass
    def __mod__(self, other:Any)->Any:
        if isinstance(other,int) or isinstance(other,float):
            return self.__class__(self.value()%other)
        else:
            return self.value()%other.value()

    def __round__(self, ndigits:int=0)->Self:
        return self.__class__(self.value().__round__(ndigits))

    
    def __abs__(self)->Self:
        return self.__class__(self.value().__abs__())
    def __neg__(self)->Self:
        return self.__class__(self.value().__neg__())
    def __pos__(self)->Self:
        return self.__class__(self.value().__pos__())
    def __trunc__(self)->Self:
        return self.__class__(self.value().__trunc__())
    def __ceil__(self)->Self:
        return self.__class__(self.value().__ceil__())
    def __floor__(self)->Self:
        return self.__class__(self.value().__floor__())
    def __sizeof__(self)->int:
        return self.value().__sizeof__()
    def __hash__(self)->int:
        return self.value().__hash__()
    def __bool__(self)->bool:
        return float(self.value()).__bool__()
    def is_integer(self)->bool:
        return float(self.value()).is_integer()
    def __gt__(self,other:Self)->bool:
        return self.value().__gt__(other.value())
    def __lt__(self,other:Self)->bool:
        return self.value().__lt__(other.value())
    def __le__(self,other:Self)->bool:
        return self.value().__le__(other.value())
    def __ge__(self,other:Self)->bool:
        return self.value().__ge__(other.value())
    def __add__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__add__(other.value()))
    def __sub__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__sub__(other.value()))
    def as_integer_ratio(self)->Tuple[float,float]:
        return self.value().as_integer_ratio()

    pass


class Str(Generic[Tag]):
    __value:str
    def __init__(self, value:str):
        self.__value=value
        pass

    def value(self)->str:
        return self.__value

    # note the following type: ignore is to get the desired behaviour from mypy --strict-equality
    # (as at mypy 1.3.0) i.e. forbid comparison of Int[XTag] and Int[YTag], noting:
    # - python itself insists on being able to compare values of unrelated type (for example
    #   to implement x in y)
    # - the type: ignore avoid mypy error for non-liskoff substitutability (because
    #   python's Object.__eq__ has other as Any)
    # - with other: Any, mypy --strict-equality gives no error at all
    # - magically (incorrectly?), the type: ignore does not suppress checks on type of other, but
    #   does suppress the substitution error
    def __eq__(self,other:Self)->bool:  # type: ignore
        '''equality test, self and other have the same type and same value'''
        '''i.e. recommend stick to using Int[X] like:
              class Hours(Int[HoursTag]):pass
           ... and not inherit from Hours.
           If you choose to inherit from Hours, make sure you write your own __eq__'''
        if type(other) is not type(self):
            return False
        return self.value().__eq__(other.value())

    def __ne__(self,other:Self)->bool:  # type: ignore
        return not self.__eq__(other)

    def __str__(self)->str:
        return str(self.value())

    def __repr__(self)->str:
        return repr(self.value())

    def __format__(self, format_spec:str)->str:
        return self.value().__format__(format_spec)

    def splitlines(self,keepends:bool=False)->List[Self]:
        return [self.__class__(_) for _ in self.value().splitlines()]

    def encode(self,encoding:str='utf-8', errors:str='strict')->bytes:
        return self.value().encode()

    def __contains__(self,other:str)->bool:
        return self.value().__contains__(other)

    def zfill(self,width:int)->Self:
        return self.__class__(self.value().zfill(width))

    def format_map(self,mapping:Mapping[Any,Any]) -> Self:
        return self.__class__(self.value().format_map(mapping))

    def format(self,*args:Any,**kwargs:Any) -> Any:
        return self.__class__(self.value().format(*args,**kwargs))
    
    def expandtabs(self,tabsize:int=8)->Self:
        return self.__class__(self.value().expandtabs(tabsize))

    def __getitem__(self,key:Any) -> Any:
        return self.value().__getitem__(key)

    
    def capitalize(self)->Self:
        return self.__class__(self.value().capitalize())
    def lower(self)->Self:
        return self.__class__(self.value().lower())
    def swapcase(self)->Self:
        return self.__class__(self.value().swapcase())
    def title(self)->Self:
        return self.__class__(self.value().title())
    def casefold(self)->Self:
        return self.__class__(self.value().casefold())
    def upper(self)->Self:
        return self.__class__(self.value().upper())
    def __len__(self)->int:
        return self.value().__len__()
    def __sizeof__(self)->int:
        return self.value().__sizeof__()
    def __hash__(self)->int:
        return self.value().__hash__()
    def isalnum(self)->bool:
        return self.value().isalnum()
    def isdecimal(self)->bool:
        return self.value().isdecimal()
    def isidentifier(self)->bool:
        return self.value().isidentifier()
    def isprintable(self)->bool:
        return self.value().isprintable()
    def isascii(self)->bool:
        return self.value().isascii()
    def islower(self)->bool:
        return self.value().islower()
    def isnumeric(self)->bool:
        return self.value().isnumeric()
    def isspace(self)->bool:
        return self.value().isspace()
    def isupper(self)->bool:
        return self.value().isupper()
    def isalpha(self)->bool:
        return self.value().isalpha()
    def isdigit(self)->bool:
        return self.value().isdigit()
    def istitle(self)->bool:
        return self.value().istitle()
    def __mul__(self,n:int)->Self:
        return self.__class__(self.value().__mul__(n))
    def __gt__(self,other:Self)->bool:
        return self.value().__gt__(other.value())
    def __lt__(self,other:Self)->bool:
        return self.value().__lt__(other.value())
    def __le__(self,other:Self)->bool:
        return self.value().__le__(other.value())
    def __ge__(self,other:Self)->bool:
        return self.value().__ge__(other.value())
    def __add__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__add__(other.value()))
    @overload
    def rfind(self, sub:str) -> int:
        pass
    @overload
    def rfind(self, sub:str, start:int)->int:
        pass
    @overload
    def rfind(self, sub:str, start:int, end:int)->int:
        pass
    def rfind(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().rfind(*args)
    @overload
    def find(self, sub:str) -> int:
        pass
    @overload
    def find(self, sub:str, start:int)->int:
        pass
    @overload
    def find(self, sub:str, start:int, end:int)->int:
        pass
    def find(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().find(*args)
    @overload
    def rindex(self, sub:str) -> int:
        pass
    @overload
    def rindex(self, sub:str, start:int)->int:
        pass
    @overload
    def rindex(self, sub:str, start:int, end:int)->int:
        pass
    def rindex(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().rindex(*args)
    @overload
    def index(self, sub:str) -> int:
        pass
    @overload
    def index(self, sub:str, start:int)->int:
        pass
    @overload
    def index(self, sub:str, start:int, end:int)->int:
        pass
    def index(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().index(*args)
    @overload
    def count(self, sub:str) -> int:
        pass
    @overload
    def count(self, sub:str, start:int)->int:
        pass
    @overload
    def count(self, sub:str, start:int, end:int)->int:
        pass
    def count(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().count(*args)
    @overload
    def translate(self, sub:str) -> int:
        pass
    @overload
    def translate(self, sub:str, start:int)->int:
        pass
    @overload
    def translate(self, sub:str, start:int, end:int)->int:
        pass
    def translate(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().translate(*args)
    @overload
    def endswith(self, s:str) -> bool:
        pass
    @overload
    def endswith(self, s:str, start:int)->bool:
        pass
    @overload
    def endswith(self, s:str, start:int, end:int)->bool:
        pass
    def endswith(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().endswith(*args)
    @overload
    def startswith(self, s:str) -> bool:
        pass
    @overload
    def startswith(self, s:str, start:int)->bool:
        pass
    @overload
    def startswith(self, s:str, start:int, end:int)->bool:
        pass
    def startswith(self, *args:Any, **kwargs:Any) -> Any:
        return self.value().startswith(*args)
    def strip(self, chars:str|None=None)->Self:
        return self.__class__(self.value().strip(chars))
    def lstrip(self, chars:str|None=None)->Self:
        return self.__class__(self.value().lstrip(chars))
    def rstrip(self, chars:str|None=None)->Self:
        return self.__class__(self.value().rstrip(chars))
    def replace(self, old:str, new:str, count:int=-1)->Self:
        return self.__class__(self.value().replace(old,new,count))
    def split(self, sep:Optional[str]=None, max_split:int=-1)->List[str]:
        return self.value().split(sep,max_split)
    def rsplit(self, sep:Optional[str]=None, max_split:int=-1)->List[str]:
        return self.value().rsplit(sep,max_split)
    def partition(self,sep:str) -> Tuple[str,str,str]:
        return self.value().partition(sep)
    def rpartition(self,sep:str) -> Tuple[str,str,str]:
        return self.value().rpartition(sep)
    def removeprefix(self,sub:str)->Self:
        return self.__class__(self.value().removeprefix(sub))
    def removesuffix(self,sub:str)->Self:
        return self.__class__(self.value().removesuffix(sub))
    def center(self,width:int,fillchar:str=' ')->Self:
        return self.__class__(self.value().center(width,fillchar))
    def ljust(self,width:int,fillchar:str=' ')->Self:
        return self.__class__(self.value().ljust(width,fillchar))
    def rjust(self,width:int,fillchar:str=' ')->Self:
        return self.__class__(self.value().rjust(width,fillchar))

    pass


class Bool(Generic[Tag]):
    __value:bool

    def __init__(self, value:bool):
        self.__value=value
        pass

    def value(self)->bool:
        return self.__value

    # note the following type: ignore is to get the desired behaviour from mypy --strict-equality
    # (as at mypy 1.3.0) i.e. forbid comparison of Bool[XTag] and Bool[YTag], noting:
    # - python itself insists on being able to compare values of unrelated type (for example
    #   to implement x in y)
    # - the type: ignore avoid mypy error for non-liskoff substitutability (because
    #   python's Object.__eq__ has other as Any)
    # - with other: Any, mypy --strict-equality gives no error at all
    # - magically (incorrectly?), the type: ignore does not suppress checks on type of other, but
    #   does suppress the substitution error
    def __eq__(self,other:Self)->bool:  # type: ignore
        '''equality test, self and other have the same type and same value'''
        '''i.e. recommend stick to using Bool[X] like:
              class Hours(Bool[HoursTag]):pass
           ... and not inherit from Hours.
           If you choose to inherit from Hours, make sure you write your own __eq__'''
        if type(other) is not type(self):
            return False
        return self.value().__eq__(other.value())

    def __ne__(self,other:Self)->bool:  # type: ignore
        return not self.__eq__(other)

    def __str__(self)->str:
        return str(self.value())

    def __bool__(self)->bool:
        return self.value()

    def __repr__(self)->str:
        return repr(self.value())

    def __format__(self, format_spec:str)->str:
        return self.value().__format__(format_spec)

    def is_integer(self)->bool:
        return True  # that's what python method does

    
    def __sizeof__(self)->int:
        return self.value().__sizeof__()
    def __hash__(self)->int:
        return self.value().__hash__()
    def __gt__(self,other:Self)->bool:
        return self.value().__gt__(other.value())
    def __lt__(self,other:Self)->bool:
        return self.value().__lt__(other.value())
    def __le__(self,other:Self)->bool:
        return self.value().__le__(other.value())
    def __ge__(self,other:Self)->bool:
        return self.value().__ge__(other.value())
    def __and__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__and__(other.value()))
    def __or__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__or__(other.value()))
    def __xor__(self,other:Self)->Self:
        if type(other) is not type(self):
            return NotImplemented
        return self.__class__(self.value().__xor__(other.value()))

    pass




# coding: utf-8

# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# Assert reports values of both terms where assertion fails, without trying
# to be fancy.
#
import re
from typing import Union

class Assert():
    def __init__(self,x):
        self.x=x
        pass
    def isInstanceOf(self,y):
        if not isinstance(self.x,y):
            xc=self.x.__class__.__name__
            yc=y.__name__
            raise Exception(f'{self.x!r} is not an instance of class {yc} (it is of class {xc})')
        return self.x
    def isNotInstanceOf(self,y):
        if isinstance(self.x,y):
            xc=self.x.__class__.__name__
            yc=y.__name__
            raise Exception(f'{self.x!r} is unexpectedly an instance of class {yc} (it is of class {xc})')
        return self.x
    def isSubclassOf(self,y):
        if not issubclass(self.x,y):
            yc=y.__name__
            raise Exception(f'{self.x!r} is not an subclass of class {yc}')
        return self.x
    def isNotSubclassOf(self,y):
        if issubclass(self.x,y):
            yc=y.__name__
            raise Exception(f'{self.x!r} is unexpectedly an subclass of class {yc}')
        return self.x
    def hasAttr(self,y:str):
        if not hasattr(self.x,y):
            raise Exception(f'{self.x!r} has no {y!r} attribute')
    def doesNotHaveAttr(self,y:str):
        if hasattr(self.x,y):
            raise Exception(f'{self.x!r} unexpectedly has {y!r} attribute')
    def startsWith(self,y):
        if not self.x[0:len(y)]==y:
            raise Exception(f'{self.x!r} does not start with {y!r}')
        return self.x
    def startswith(self,y):
        return self.startsWith(y)
    def starts_with(self,y):
        return self.startsWith(y)
    def endsWith(self,y):
        if not self.x[-len(y):]==y:
            raise Exception(f'{self.x!r} does not end with {y!r}')
        return self.x
    def endswith(self,y):
        return self.endsWith(y)
    def ends_with(self,y):
        return self.endsWith(y)
    def contains(self,y):
        if not y in self.x:
            raise Exception(f'{self.x!r} does not contain {y!r}')
        return self.x
    def isNotIn(self,y):
        if self.x in y:
            raise Exception(f'{self.x!r} is unexpectedly in {y!r}')
        pass
    def isIn(self,y):
        if not self.x in y:
            raise Exception(f'{self.x!r} is not in {y!r}')
        pass
    def matches(self,x:Union[str,re.Pattern]):
        c:re.Pattern
        pattern:str
        if isinstance(x,re.Pattern):
            pattern=x.pattern
            c=x
        elif isinstance(x,str):
            pattern=x
            c=re.compile(pattern)
        else:
            assert False, '?'
        if not c.match(self.x):
            raise Exception(f'{self.x!r} does not match regular expression {pattern!r}')
        pass
    def doesNotMatch(self,x:Union[str,re.Pattern]):
        c:re.Pattern
        pattern:str
        if isinstance(x,re.Pattern):
            pattern=x.pattern
            c=x
        elif isinstance(x,str):
            pattern=x
            c=re.compile(pattern)
        else:
            assert False, '?'
        if c.match(self.x):
            raise Exception(f'{self.x!r} unexpectedly matches regular expression {pattern!r}')
        pass
    def __lt__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if not self.x < y:
            raise Exception(f'{self.x!r} is not less than {y!r}')
        return self.x
    def __le__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if not self.x <= y:
            raise Exception(f'{self.x!r} is not less than or equal to {y!r}')
        return self.x
    def __eq__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if not self.x == y:
            raise Exception(f'{self.x!r} is not equal to {y!r}')
        return self.x
    def __ne__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if self.x == y:
            raise Exception(f'{self.x!r} is unexpectedly equal to {y!r}')
        return self.x
    def __gt__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if not self.x > y:
            raise Exception(f'{self.x!r} is not greater than {y!r}')
        return self.x
    def __ge__(self,y):
        if type(y) is not type(self.x):
            xt=type(self.x)
            yt=type(y)
            raise Exception(f'{y} is a {yt}, not a {xt} (which is what {self.x} is)')
        if not self.x >= y:
            raise Exception(f'{self.x!r} is not greater than or equal to {y!r}')
        return self.x
    def is_(self,y):
        if not self.x is y:
            raise Exception(f'{self.x} is not the same object as {y}')
        return self.x
    def isNot(self,y):
        if self.x is y:
            raise Exception(f'{self.x} is the same object as {y}')
        return self.x
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
from xju.newtype import Int,Float,Union
from typing import overload, Self
import time
import asyncio
from dataclasses import dataclass
from typing import Sequence

class HoursTag:pass
class MinutesTag:pass
class SecondsTag:pass

class Hours(Int[HoursTag]): pass
class Minutes(Int[MinutesTag]): pass
class Seconds(Int[SecondsTag]): pass

class DurationTag:pass
class Duration(Float[DurationTag]):pass


@dataclass(order=True)
class Timestamp():
    __value:float
    def __init__(self, value:float|int):
        self.__value=float(value)
    def value(self):
        return self.__value
    def __str__(self):
        return str(self.__value)
    def __repr__(self):
        return repr(self.__value)
    def __format__(self,f:str)->str:
        return self.__value.__format__(f)
    def __add__(self, x:Union[Duration,Hours,Minutes,Seconds]):
        if isinstance(x,Hours):
            return Timestamp(self.value()+x.value()*3600)
        if isinstance(x,Minutes):
            return Timestamp(self.value()+x.value()*60)
        if isinstance(x,Seconds):
            return Timestamp(self.value()+x.value())
        return Timestamp(self.value()+x.value())
    def __radd__(self, duration:Union[Duration,Hours,Minutes,Seconds]):
        return self.__add__(duration)

    @overload
    def __sub__(self, t2:Self)->Duration:
        pass
    @overload
    def __sub__(self, duration:Union[Duration,Hours,Minutes,Seconds])->Self:
        pass
    def __sub__(self, x):
        if isinstance(x,Duration):
            return Timestamp(self.value()-x.value())
        if isinstance(x,Hours):
            return Timestamp(self.value()-x.value()*3600)
        if isinstance(x,Minutes):
            return Timestamp(self.value()-x.value()*60)
        if isinstance(x,Seconds):
            return Timestamp(self.value()-x.value())
        if isinstance(x,Timestamp):
            return Duration(self.value()-x.value())
        x_type=type(x)
        assert False, f'cannot subtract {x} of type {x_type} from Timestamp'
    def __float__(self):
        return self.__value
    pass

def now()->Timestamp:
    return Timestamp(time.time())

def sleep_for(x:Duration):
    time.sleep(x.value())
    pass

async def async_sleep_until(deadline: Timestamp) -> None:
    """sleep while we're not yet passed {deadline}"""
    await asyncio.sleep(max(0.0, (deadline-now()).value()))

async def async_sleep_until_first_of(a_deadline: Timestamp,
                                     *more_deadlines: Timestamp) -> None:
    deadlines:list[Timestamp]=[a_deadline]+list(more_deadlines)
    deadlines.sort()
    await async_sleep_until(deadlines.pop(0))



# coding: utf-8
#
# Copyright (c) 2023 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# json schema represented as a python type var, e.g.:
#
#   @dataclasses.dataclass
#   class Address:
#     number: str|int
#     street: str
#
#   codec(Address).decode({'number':32, 'street': 'asler'})==Address(32,'asler')
#   codec(Address).decode({'number':'42 biz', 'street': 'asler'})==Address('42 biz','asler')
#   codec(Address).encode(Address(32,'asler'))=={'number':32, 'street': 'asler'}
#
# For examples see json_codec.py.test
#
from dataclasses import dataclass
from xju.xn import Xn,in_context,in_function_context,readable_repr, AllFailed
from typing import TypeVar, Generic, Type, cast, Any, Protocol, Self, Callable, get_type_hints
from typing import Sequence, Literal, NewType
from typing import _LiteralGenericAlias  # type: ignore  # mypy 1.1.1
from typing import _UnionGenericAlias  # type: ignore  # mypy 1.1.1
from typing import _GenericAlias  # type: ignore  # mypy 1.2.0
from typing import get_origin,get_args, runtime_checkable
from types import GenericAlias, UnionType, NoneType
import xju.newtype
from enum import Enum, EnumType, EnumMeta
import json

T=TypeVar('T')

JsonType = None|bool|dict|list|float|str

class TypeScriptUQNTag:pass
class TypeScriptUQN(xju.newtype.Str[TypeScriptUQNTag]):pass

class TypeScriptSourceCodeTag:pass
class TypeScriptSourceCode(xju.newtype.Str[TypeScriptSourceCodeTag]):pass

@dataclass
class TypeScriptNamespace:
    defs: dict[TypeScriptUQN, TypeScriptSourceCode | Self]

    def get_namespace_of(self,fqn:Sequence[TypeScriptUQN]) -> Self:
        match len(fqn):
            case 1:
                return self
            case _:
                m=self.defs.setdefault(fqn[0],self.__class__({}))
                assert isinstance(m,self.__class__)
                return m.get_namespace_of(fqn[1:])
        pass

    def get_formatted_defs(self, export:Literal['']|Literal['export ']='')->TypeScriptSourceCode:
        '''get defs of this namespace as (possibly multi-level) namespace content
           - top level defs are exported or not per {export}
           - sub-level defs are always exported'''
        result:list[TypeScriptSourceCode]=[]
        for name, d in self.defs.items():
            match d:
                case TypeScriptSourceCode():
                    result.append(TypeScriptSourceCode(f'{export}{d}'))
                case TypeScriptNamespace:
                    result.append(TypeScriptSourceCode(
                        f"{export}namespace {name} {{\n"
                        f"    {indent(4, d.get_formatted_defs(export='export '))}\n"
                        f"}}\n"))
            pass
        return TypeScriptSourceCode('\n'.join([_.value() for _ in result]))
    pass

@dataclass
class TypeScriptBackRefs:
    type_back_ref:Callable[[],TypeScriptSourceCode]
    isa_back_ref:Callable[[TypeScriptSourceCode],TypeScriptSourceCode]
    asa_back_ref:Callable[[TypeScriptSourceCode],TypeScriptSourceCode]


class CodecProto(Generic[T], Protocol):
    def encode(self,x:T) -> JsonType:
        'encode {self.t} {x} to json'
        pass
    def decode(self,x:JsonType) -> T:
        'decode {x} to a {self.t}'
        pass
    def get_json_schema(self) -> dict:
        "get jsonschema matching T's encoded form"
        pass
    def typescript_type(self) -> TypeScriptUQN:
        '''return typescript unqualified equivalent type for T'''
        pass
    def ensure_typescript_defs(self, namespace:TypeScriptNamespace) -> None:
        '''ensure namespace {namespace} contains {self} typescript defs'''
        pass
    def add_typescript_alias(self, namespace:TypeScriptNamespace, fqn:Sequence[TypeScriptUQN]) -> None:
        '''add {fqn} as typescript alias for {self.t}'''
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''get typescript "{expression} is a T" code, adding any necessary definitions to {namespace}
           result is a type-guard expression, examples of some possible results:
             parameter       possible code                  note
             "x.y"           "(typeof (x.y) == 'number')"   (uses built-in type guard pattern)
             "z"             "m.isInstanceOfFred(z)"        (calls a type-guard function)'''
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''get typescript "{expression} as a T" code, adding any necessary definitions to {namespace}
           result is a cast expression that throws an Error if {expression} is not a T
        '''
        
def codec(t: Type[T]) -> CodecProto[T]:
    '''build codec to encode/decode a "t" to/from json

       - note requires json_codec_mypy_plugin for proper typing and no errors with mypy
    '''
    return Codec[T](t)


# implementation of CodecProto - use codec above with json_codec_mypy_plugin
# rather than using this class directly
class Codec(Generic[T]):
    t:Type[T]
    def __init__(self, t: Type[T]):
        'initialse json decoder for type %(t)r'
        self.t=t
        self.codec:CodecImplProto=_explodeSchema(self.t, {})
        pass

    def __repr__(self):
        return f'{self.t!r} json codec'

    def encode(self,x:T) -> JsonType:
        'encode {self.t} {x} to json'
        return self.codec.encode(x,None)

    def decode(self,x:JsonType) -> T:
        'decode {x} to a {self.t}'
        try:
            return cast(T, self.codec.decode(x,None))
        except Exception:
            raise in_function_context(Codec.decode,vars()) from None
        pass

    def get_json_schema(self) -> dict:
        definitions: dict[str,dict] = {}
        result = self.codec.get_json_schema(definitions, None)
        result.update({
            "$id": "https://example.com/address.schema.json",
            "$schema": "https://json-schema.org/draft/2020-12/schema"})
        result.update({
            'definitions':definitions
        })
        return result
    def typescript_type(self) -> TypeScriptUQN:
        '''return typescript unqualified equivalent type for T'''
        return self.codec.typescript_type(None)
    def ensure_typescript_defs(self, namespace:TypeScriptNamespace) -> None:
        return self.codec.ensure_typescript_defs(namespace)
    def add_typescript_alias(self, namespace:TypeScriptNamespace, fqn:Sequence[TypeScriptUQN]) -> None:
        '''add {fqn} as typescript alias for {self.t}'''
        assert len(fqn)>0
        try:
            target_namespace=namespace.get_namespace_of(fqn)
            if fqn[-1] in target_namespace.defs:
                m='.'.join([str(_) for _ in fqn])
                raise Exception(
                    f'{m} is already defined as {target_namespace.defs[fqn[-1]]}')
            target_namespace.defs[fqn[-1]]=TypeScriptSourceCode(
                f'type {fqn[-1]} = {self.typescript_type()};')
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{fqn[-1]}")] = TypeScriptSourceCode(
                f"function asInstanceOf{fqn[-1]}(v:any):{fqn[-1]} {{ return {self.codec.get_typescript_asa(TypeScriptSourceCode('v'),namespace,None)}; }}")
            target_namespace.defs[TypeScriptUQN(f"isInstanceOf{fqn[-1]}")] = TypeScriptSourceCode(
                f"function isInstanceOf{fqn[-1]}(v:any):v is {fqn[-1]} {{ return {self.codec.get_typescript_isa(TypeScriptSourceCode('v'),namespace,None)}; }}")
        except:
            raise in_function_context(Codec.add_typescript_alias,vars())
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''get typescript "{expression} is a T" code, adding any necessary definitions to {namespace}
           result is a type-guard expression, examples of some possible results:
             parameter       possible code                  note
             "x.y"           "(typeof (x.y) == 'number')"   (uses built-in type guard pattern)
             "z"             "m.isInstanceOfFred(z)"        (calls a type-guard function)'''
        return self.codec.get_typescript_isa(expression,namespace,None)
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''get typescript "{expression} as a T" code, adding any necessary definitions to {namespace}
           result is a cast expression that throws an Error if {expression} is not a T
        '''
        return self.codec.get_typescript_asa(expression,namespace,None)
    pass

class CodecImplProto(Protocol):
    def encode(self,x:Any,back_ref:None|Callable[[Any],JsonType])->JsonType:
        pass
    def decode(self,x:JsonType,back_ref:None|Callable[[JsonType],Any])->Any:
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        pass
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> TypeScriptUQN:
        pass
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        pass
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        pass
    pass

Atom=TypeVar('Atom',int,str,bool,float)

NewInt=TypeVar('NewInt',bound=xju.newtype.Int)
NewFloat=TypeVar('NewFloat',bound=xju.newtype.Float)
NewStr=TypeVar('NewStr',bound=xju.newtype.Str)

class NoopCodecImpl(Generic[Atom]):
    t:Type[Atom]
    def __init__(self, t:Type[Atom]):
        self.t=t
    def encode(self, x:Atom, _:None|Callable[[Any],JsonType])->Atom:
        if type(x) is not self.t and not (
                self.t is float and type(x) is int):
            raise Exception(f'{x!r} is not a {self.t}')
        return x
    def decode(self, x:Atom,_:None|Callable[[JsonType],Any])->Atom:
        if type(x) is not self.t and not (
                self.t is float and type(x) is int):
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not a {self.t}')
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        st: str
        if self.t is int: st='integer'
        if self.t is str: st='string'
        if self.t is float: st='number'
        if self.t is bool: st='boolean'
        return { 'type': st }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        '''return typescript equivalent type for T'''
        if self.t is int: return 'number'
        if self.t is str: return 'string'
        if self.t is float: return 'number'
        assert self.t is bool, self.t
        return 'boolean'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(typeof ({expression}) == '{self.typescript_type(back_refs)}')")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (typeof v !== '{tt}') throw new Error(`${{v}} is not a {tt} it is a ${{typeof v}}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class NoneCodecImpl:
    def encode(self, x:None,_:None|Callable[[Any],JsonType])->None:
        if x is not None:
            raise Exception(f'{x!r} is not None')
        return x
    def decode(self, x:None,_:None|Callable[[JsonType],Any])->None:
        if x is not None:
            raise Exception(f'{x!r} is not None')
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return { 'type': 'null' }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        '''return typescript equivalent type for None'''
        return 'null'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(({expression}) === {self.typescript_type(back_refs)})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"((v: any): null => {{\n"
            f"    if (v !== null) throw new Error(`${{v}} is not null it is a ${{typeof v}}`);\n"
            f"    return v as null;\n"
            f"}})({expression})")
    pass

class ListCodecImpl:
    def __init__(self, value_codec:CodecImplProto):
        self.value_codec=value_codec
        pass
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.encode(_,back_ref) for _ in x]
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.decode(_,back_ref) for _ in x]
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "type": "array",
            "items": self.value_codec.get_json_schema(definitions, self_ref)
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return f"Array<{self.value_codec.typescript_type(back_refs)}>"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"(((v:any):v is {tt}=>(Array.isArray(v) && "
            f"    v.filter((x)=>(\n"
            f"        !{indent(8,self.value_codec.get_typescript_isa(TypeScriptSourceCode('x'),namespace,back_refs))})).length==0))({expression}))")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{try{{\n"
            f"    if (!Array.isArray(v)) throw new Error(`${{v}} is not an array it is a ${{typeof v}}`);\n"
            f"    v.forEach((x)=>{indent(4,self.value_codec.get_typescript_asa(TypeScriptSourceCode('x'),namespace,back_refs))});\n"
            f"    return v as {tt};\n"
            f"}}catch(e:any){{throw new Error(`${{v}} is not a {tt} because ${{e}}`);}}}})({expression})")
    pass

class AnyListCodecImpl:
    def __init__(self):
        self.value_codec=AnyJsonCodecImpl()
        pass
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.encode(_,back_ref) for _ in x]
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.decode(_,back_ref) for _ in x]
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "type": "array"
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return f"Array<any>"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(Array.isArray({expression}))")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (!Array.isArray(v)) throw new Error(`${{v}} is not an Array it is a ${{typeof v}}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class TupleCodec:
    def __init__(self,value_codecs):
        self.value_codecs=value_codecs
        self.number_of_codecs=len(value_codecs)
        pass
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]) -> list:
        'encode tuple {x} as a {self.number_of_codecs}-element list'
        try:
            if type(x) is not tuple:
                raise Exception(f'{x!r} is not a tuple')
            if len(x) != self.number_of_codecs:
                l=len(x)
                raise Exception(f'{x} does not have {self.number_of_codecs} items (it has {l} items)')
            return [c.encode(v,back_ref) for c,v in zip(self.value_codecs,x)]
        except Exception:
            raise in_function_context(TupleCodec.encode,vars()) from None
        pass
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]) -> tuple:
        try:
            if type(x) is not list:
                raise Exception(f'{x!r} is not a list')
            if len(x) != self.number_of_codecs:
                l=len(x)
                raise Exception(f'{x} does not have {self.number_of_codecs} items (it has {l} items)')
            result=[c.decode(v,back_ref) for c,v in zip(self.value_codecs,x)]
            return tuple(result)
        except Exception:
            raise in_function_context(TupleCodec.decode,vars()) from None
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                codec.get_json_schema(definitions, self_ref) for codec in self.value_codecs
            ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        value_types=','.join([_.typescript_type(back_refs) for _ in self.value_codecs])
        return f"[{value_types}]"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        isas=[' &&\n            '+indent(12,self.value_codecs[i].get_typescript_isa(f"v[{i}]",namespace,back_refs))
              for i in range(0, self.number_of_codecs)]
        return TypeScriptSourceCode(
            f"((v:any): v is {self.typescript_type(back_refs)}=>{{\n"
            f"    if (Array.isArray(v)) {{\n" +
            f"        return v.length == {self.number_of_codecs}" +
            f"".join(isas) + ";\n"+
            f"    }}\n"+
            f"    else return false;\n"+
            f"}})({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        asas=['    '+indent(4,self.value_codecs[i].get_typescript_asa(f"v[{i}]",namespace,back_refs))+';\n'
              for i in range(0, self.number_of_codecs)]
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{try{{\n" +
            f"    if (!Array.isArray(v)) throw new Error(`${{v}} is not an array it is a ${{typeof v}}`);\n" +
            f"    if (v.length != {self.number_of_codecs}) throw new Error(`${{v}} does not have {self.number_of_codecs} elements (it has ${{v.length}} elements)`);\n" +
            ''.join(asas)+
            f"    return v as {tt};\n"+
            f"}}catch(e:any){{ throw new Error(`${{v}} is not a {tt} because ${{e}}`);}}}})({expression})")
    pass

class UnionCodecImpl:
    def __init__(self,allowed_types, value_codecs):
        self.allowed_types=allowed_types
        self.value_codecs=value_codecs
        pass
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]) -> JsonType:
        'encode {x!r} as one of {self.allowed_types}'
        try:
            exceptions=list[BaseException]()
            for t, c in self.value_codecs.items():
                try:
                    return c.encode(x,back_ref)
                except Exception as e:
                    exceptions.append(in_context(f'decode as {t!r}'))
                    pass
                pass
            raise AllFailed(exceptions)
        except Exception:
            raise in_function_context(UnionCodecImpl.encode,vars()) from None
        pass
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]) -> tuple:
        '''decode {x!r} as one of {self.allowed_types}'''
        try:
            exceptions=list[BaseException]()
            for t, c in self.value_codecs.items():
                try:
                    return c.decode(x,back_ref)
                except Exception as e:
                    exceptions.append(in_context(f'decode as {t!r}'))
                    pass
                pass
            raise AllFailed(exceptions)
        except Exception:
            raise in_function_context(UnionCodecImpl.decode,vars()) from None
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "oneOf": [ codec.get_json_schema(definitions, self_ref) for codec in self.value_codecs.values() ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return '|'.join([_.typescript_type(back_refs) for _ in self.value_codecs.values()])
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        isas=[' ||\n       '+indent(12,value_codec.get_typescript_isa(f"v",namespace,back_refs))
              for value_codec in self.value_codecs.values()]
        return TypeScriptSourceCode(
            f"((v:any): v is {self.typescript_type(back_refs)}=>{{\n"
            f"    return false" +
            f"".join(isas) + ";\n"+
            f"}})({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        asas=[f"    try{{\n"+
              f"        {indent(8,value_codec.get_typescript_asa('v',namespace,back_refs))};\n"+
              f"        return v as {tt};\n"+
              f"    }}\n"+
              f"    catch(e:any){{\n"+
              f"        es.push(e.message);\n"+
              f"    }};\n"
              for value_codec in self.value_codecs.values()]
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{try{{\n" +
            f"    var es = new Array<string>();\n"+
            "".join(asas)+
            f"    throw new Error(es.join(' and '));\n"+
            f"}}catch(e:any)\n"+
            f"{{\n"+
            f"    throw new Error(`${{v}} is not a {tt} because ${{e}}`);\n"+
            f"}}}})({expression})")
    pass

class DictCodecImpl:
    def __init__(self, key_codec, value_codec):
        self.key_codec=key_codec
        self.value_codec=value_codec
    def encode(self,x:dict,back_ref:None|Callable[[Any],JsonType])->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        result:dict={}
        for k,v in x.items():
            ek,ev=self.key_codec.encode(k,back_ref),self.value_codec.encode(v,back_ref)
            if not isinstance(ek, str):
                raise Exception(f'key encoder {self.key_codec} produced non-str {ek} from dict key {k}')
            result[ek]=ev
            pass
        return result
    def decode(self,x,back_ref:None|Callable[[JsonType],Any])->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        return {self.key_codec.decode(k,back_ref):self.value_codec.decode(v,back_ref)
                for k,v in x.items()}
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'type': 'object',
            'additionalProperties': self.value_codec.get_json_schema(definitions, self_ref)
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return f"{{ [key: {self.key_codec.typescript_type(back_refs)}]: {self.value_codec.typescript_type(back_refs)} }}"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(((x:any):x is {self.typescript_type(back_refs)}=>(\n"
            f"    x !== null &&\n"
            f"    typeof (x) === 'object' &&\n"
            f"    !Array.isArray(x) &&\n"
            f"    (():boolean=>{{for (const k in x){{\n"
            f"        if (!x.hasOwnProperty(k)) continue;\n"
            f"        if (!({indent(8,self.key_codec.get_typescript_isa(TypeScriptSourceCode('k'),namespace,back_refs))} )||\n"
            f"            !({indent(8,self.value_codec.get_typescript_isa(TypeScriptSourceCode('x[k]'),namespace,back_refs))})) return false;}}\n"
            f"    return true;\n"
            f"}})()))({expression}))")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((x: any): {tt} => {{try{{\n"
            f"    if (x === null) throw new Error(`${{x}} is not an object it is null`);\n"
            f"    if (typeof x !== 'object') throw new Error(`${{x}} is not an object it is a ${{typeof x}}`);\n"
            f"    if (Array.isArray(x)) throw new Error(`${{x}} is not an object it is an array`);\n"
            f"    for(const k in x){{\ntry{{\n"+
            f"        if (!x.hasOwnProperty(k)) continue;\n"+
            f"        const v=x[k];\n"
            f"        {indent(8,self.key_codec.get_typescript_asa(TypeScriptSourceCode('k'),namespace,back_refs))};\n"
            f"        {indent(8,self.value_codec.get_typescript_asa(TypeScriptSourceCode('v'),namespace,back_refs))};\n"
            f"    }}catch(e:any){{\n"
            f"        throw new Error(`element ${{k}} is invalid because ${{e}}`);\n"
            f"    }}}};\n"
            f"    return x as {tt};\n"
            f"}}catch(e:any){{throw new Error(`${{x}} is not a {tt} because ${{e}}`);}}}})({expression})")
    pass

class AnyDictCodecImpl:
    def __init__(self):
        self.key_codec=NoopCodecImpl[str](str)
        self.value_codec=AnyJsonCodecImpl()
    def encode(self,x:dict,back_ref:None|Callable[[Any],JsonType])->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        result:dict={}
        for k,v in x.items():
            ek,ev=self.key_codec.encode(k,back_ref),self.value_codec.encode(v,back_ref)
            result[ek]=ev
            pass
        return result
    def decode(self,x,back_ref:None|Callable[[JsonType],Any])->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        return {self.key_codec.decode(k,back_ref):self.value_codec.decode(v,back_ref) for k,v in x.items()}
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'type': 'object'
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return f"{{ [key: {self.key_codec.typescript_type(back_refs)}]: {self.value_codec.typescript_type(back_refs)} }}"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(((x:any):x is {self.typescript_type(back_refs)}=>(\n"
            f"    (x !== null &&\n"
            f"    typeof (x) === 'object' &&\n"
            f"    !Array.isArray(x) &&\n"
            f"    (():boolean=>{{for (const k in x){{\n"
            f"        if (!x.hasOwnProperty(k)) continue;\n"
            f"        if (!({indent(12,self.key_codec.get_typescript_isa(TypeScriptSourceCode('k'),namespace,back_refs))} )||\n"
            f"            !({indent(12,self.value_codec.get_typescript_isa(TypeScriptSourceCode('x[k]'),namespace,back_refs))})) return false;}}\n"
            f"        return true;\n"
            f"    }})())))({expression}))")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((x: any): {tt} => {{try{{\n"
            f"    if (x === null) throw new Error(`${{x}} is not an object it is null`);\n"
            f"    if (typeof x !== 'object') throw new Error(`${{x}} is not an object it is a ${{typeof x}}`);\n"
            f"    if (Array.isArray(x)) throw new Error(`${{x}} is not an object it is an array`);\n"
            f"    for(const k in x){{\ntry{{\n"+
            f"        if (!x.hasOwnProperty(k)) continue;\n"+
            f"        const v=x[k];\n"
            f"        {indent(8,self.key_codec.get_typescript_asa(TypeScriptSourceCode('k'),namespace,back_refs))};\n"
            f"        {indent(8,self.value_codec.get_typescript_asa(TypeScriptSourceCode('v'),namespace,back_refs))};\n"
            f"    }}catch(e:any){{\n"
            f"        throw new Error(`element ${{k}} is invalid because ${{e}}`);\n"
            f"    }}}};\n"
            f"    return x as {tt};\n"
            f"}}catch(e:any){{throw new Error(`${{x}} is not a {tt} because ${{e}}`);}}}})({expression})")
    pass

class AnyJsonCodecImpl:
    def encode(self,x,_:None|Callable[[Any],JsonType]):
        # we assume x will be subsequently json.dumps()ed so defer validation to that
        return x
    def decode(self,x,_:None|Callable[[JsonType],Any]):
        # we assume x was json.loads()ed so assume it is appropriate
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "oneOf": [ { 'type': t } for t in ['null','boolean','object','array','number','string'] ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return 'any'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            "true")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"({expression})")
    pass

class NewIntCodecImpl(Generic[NewInt]):
    t:Type[NewInt]
    def __init__(self,t:Type[NewInt]):
        self.t=t
        self.base_codec=NoopCodecImpl[int](int)
    def encode(self, x:NewInt,back_ref:None|Callable[[Any],JsonType])->int:
        if not isinstance(x, xju.newtype.Int):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value(),back_ref)
    def decode(self, x:JsonType,back_ref:None|Callable[[JsonType],Any])->NewInt:
        if not isinstance(x, int):
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not an int')
        return self.t(self.base_codec.decode(x,back_ref))
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'description': self.get_type_fqn(),
            'type': 'integer'
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return 'number'
    def ensure_typescript_defs(self, namespace) -> None:
        typescript_fqn=[TypeScriptUQN(_) for _ in self.get_type_fqn().split('.')]
        target_namespace=namespace.get_namespace_of(typescript_fqn)
        typescript_type_name=typescript_fqn[-1]
        if typescript_type_name not in target_namespace.defs:
            target_namespace.defs[typescript_type_name]=TypeScriptSourceCode(
                f"type {typescript_type_name} = number;")
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function asInstanceOf{typescript_type_name}(v: any): {typescript_type_name}\n"
                f"{{\n"
                f"    try{{\n"
                f"        if (typeof v !== 'number') throw new Error(`${{v}} is a ${{typeof v}}`);\n"
                f"        return v;\n"
                f"    }}\n"
                f"    catch(e:any){{\n"
                f"        throw new Error(`${{v}} is not a {typescript_type_name} because ${{e}}`);\n"
                f"    }}\n"
                f"}}")
            target_namespace.defs[TypeScriptUQN(f'isInstanceOf{typescript_type_name}')]=TypeScriptSourceCode(
                f"function isInstanceOf{typescript_type_name}(v:any): v is number\n"
                f"{{\n"
                f"    return typeof v === 'number';\n"
                f"}}")
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(typeof ({expression}) == '{self.typescript_type(back_refs)}')")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (typeof v !== '{tt}') throw new Error(`${{v}} is not a {self.get_type_fqn()} i.e. a {tt}, it is a ${{typeof v}}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class NewFloatCodecImpl(Generic[NewFloat]):
    t:Type[NewFloat]
    def __init__(self,t:Type[NewFloat]):
        self.t=t
        self.base_codec=NoopCodecImpl[float](float)
    def encode(self, x:NewFloat,back_ref:None|Callable[[Any],JsonType])->float:
        if not isinstance(x, xju.newtype.Float):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value(),back_ref)
    def decode(self, x:JsonType,back_ref:None|Callable[[JsonType],Any])->NewFloat:
        if type(x) is not float and type(x) is not int:
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not a float (or an int)')
        return self.t(self.base_codec.decode(x,back_ref))
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'description': self.get_type_fqn(),
            'type': 'number'
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return 'number'
    def ensure_typescript_defs(self, namespace) -> None:
        typescript_fqn=[TypeScriptUQN(_) for _ in self.get_type_fqn().split('.')]
        target_namespace=namespace.get_namespace_of(typescript_fqn)
        typescript_type_name=typescript_fqn[-1]
        if typescript_type_name not in target_namespace.defs:
            target_namespace.defs[typescript_type_name]=TypeScriptSourceCode(
                f"type {typescript_type_name} = number;")
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function asInstanceOf{typescript_type_name}(v: any): {typescript_type_name}\n"
                f"{{\n"
                f"    try{{\n"
                f"        if (typeof v !== 'number') throw new Error(`${{v}} is a ${{typeof v}}`);\n"
                f"        return v;\n"
                f"    }}\n"
                f"    catch(e:any){{\n"
                f"        throw new Error(`${{v}} is not a {typescript_type_name} because ${{e}}`);\n"
                f"    }}\n"
                f"}}")
            target_namespace.defs[TypeScriptUQN(f'isInstanceOf{typescript_type_name}')]=TypeScriptSourceCode(
                f"function isInstanceOf{typescript_type_name}(v:any): v is number\n"
                f"{{\n"
                f"    return typeof v === 'number';\n"
                f"}}")
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(typeof ({expression}) == '{self.typescript_type(back_refs)}')")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (typeof v !== '{tt}') throw new Error(`${{v}} is not a {self.get_type_fqn()} i.e. a {tt}, it is a ${{typeof v}}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class NewStrCodecImpl(Generic[NewStr]):
    t:Type[NewStr]
    def __init__(self,t:Type[NewStr]):
        self.t=t
        self.base_codec=NoopCodecImpl[str](str)
    def encode(self, x:NewStr,back_ref:None|Callable[[Any],JsonType])->str:
        if not isinstance(x, xju.newtype.Str):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value(),back_ref)
    def decode(self, x:JsonType,back_ref:None|Callable[[JsonType],Any])->NewStr:
        if type(x) is not str:
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not an str')
        return self.t(self.base_codec.decode(x,back_ref))
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'description': self.get_type_fqn(),
            'type': 'string'
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return 'string'
    def ensure_typescript_defs(self, namespace) -> None:
        typescript_fqn=[TypeScriptUQN(_) for _ in self.get_type_fqn().split('.')]
        target_namespace=namespace.get_namespace_of(typescript_fqn)
        typescript_type_name=typescript_fqn[-1]
        if typescript_type_name not in target_namespace.defs:
            target_namespace.defs[typescript_type_name]=TypeScriptSourceCode(
                f"type {typescript_type_name} = string;")
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function asInstanceOf{typescript_type_name}(v: any): {typescript_type_name}\n"
                f"{{\n"
                f"    try{{\n"
                f"        if (typeof v !== 'string') throw new Error(`${{v}} is a ${{typeof v}}`);\n"
                f"        return v;\n"
                f"    }}\n"
                f"    catch(e:any){{\n"
                f"        throw new Error(`${{v}} is not a {typescript_type_name} because ${{e}}`);\n"
                f"    }}\n"
                f"}}")
            target_namespace.defs[TypeScriptUQN(f'isInstanceOf{typescript_type_name}')]=TypeScriptSourceCode(
                f"function isInstanceOf{typescript_type_name}(v:any): v is string\n"
                f"{{\n"
                f"    return typeof v === 'string';\n"
                f"}}")
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(
            f"(typeof ({expression}) == '{self.typescript_type(back_refs)}')")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (typeof v !== '{tt}') throw new Error(`${{v}} is not a {self.get_type_fqn()} i.e. a {tt}, it is a ${{typeof v}}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class LiteralStrCodecImpl:
    value:str
    def __init__(self,value:str):
        self.value=value
    def encode(self, x:str, _:None|Callable[[Any],JsonType])->str:
        if type(x) is not str or x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def decode(self, x:JsonType,_:None|Callable[[JsonType],Any])->str:
        if type(x) is not str:
            raise Exception(f'{x!r} is not a string')
        if x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'type': 'string',
            'enum': [ self.value ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return f'"{self.value}"'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v:any): v is {tt}=>(v==={tt}))({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (v !== {tt}) throw new Error(`the ${{typeof v}} ${{v}} is not the string {self.value}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class LiteralIntCodecImpl:
    value:int
    def __init__(self,value:int):
        self.value=value
    def encode(self, x:int, _:None|Callable[[Any],JsonType])->int:
        if type(x) is not int or x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def decode(self, x:JsonType,_:None|Callable[[JsonType],Any])->int:
        if type(x) is not int:
            raise Exception(f'{x!r} is not a int')
        if x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'type': 'number',
            'enum': [ self.value ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return f'{self.value}'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v:any): v is {tt}=>(v==={tt}))({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (v !== {tt}) throw new Error(`the ${{typeof v}} ${{v}} is not the number {self.value}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class LiteralBoolCodecImpl:
    value:bool
    def __init__(self,value:bool):
        self.value=value
    def encode(self, x:bool, _:None|Callable[[Any],JsonType])->bool:
        if type(x) is not bool or x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def decode(self, x:JsonType,_:None|Callable[[JsonType],Any])->bool:
        if type(x) is not bool:
            raise Exception(f'{x!r} is not a boolean')
        if x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            'type': 'boolean',
            'enum': [ self.value ]
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return 'true' if self.value else 'false'
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v:any): v is {tt}=>(v==={tt}))({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        typescript_value='true' if self.value else 'false'
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (v !== {tt}) throw new Error(`the ${{typeof v}} ${{v}} is not {self.typescript_type(None)}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass

class BytesCodecImpl:
    def __init__(self, type_var_map:dict[TypeVar,Any]|None):
        self.value_codec=_explodeSchema(list[int],type_var_map)
        pass
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]):
        if type(x) is not bytes:
            raise Exception(f'{x!r} is not of type bytes')
        return self.value_codec.encode([_ for _ in x],back_ref)
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]):
        'decode {x} as "bytes"'
        try:
            return bytes(self.value_codec.decode(x,back_ref))
        except Exception as e:
            raise in_function_context(BytesCodecImpl,vars())
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        result=self.value_codec.get_json_schema(definitions,self_ref)
        result['description']="bytes"
        return result
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return f"Array<number>"
    def ensure_typescript_defs(self, namespace) -> None:
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return self.value_codec.get_typescript_isa(expression,namespace,back_refs)
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return self.value_codec.get_typescript_asa(expression,namespace,back_refs)
    pass

@runtime_checkable
class CustomClassCodec(Protocol):
    '''implement these methods on class T to use custom encoding of instances of T'''
    @staticmethod
    def xju_json_codec_encode(
            x:object  # x is a T
    ) -> JsonType:
        assert False  #pragma NO COVER
        pass
    @staticmethod
    def xju_json_codec_decode(x:JsonType) -> object:  # must return a T
        assert False  #pragma NO COVER
        pass
    @staticmethod
    def xju_json_codec_get_json_schema(definitions:dict[str,dict]) -> dict:
        '''return json schema for T
           - may add any supporting definitions to definitions'''
        assert False  #pragma NO COVER
        return {}
    @staticmethod
    def xju_json_codec_get_typescript_isa(
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''return typescript source code that turns {expression} into a bool indicating whether the expression is a T
           - may add any supporting definitions to namespace, e.g. type for T itself'''
        assert False  #pragma NO COVER
        return TypeScriptSourceCode('')
    @staticmethod
    def xju_json_codec_get_typescript_asa(
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace) -> TypeScriptSourceCode:
        '''return typescript source code that safely casts {expression} to a T, throwing an Error if {expression} is not valid as a T
           - may add any supporting definitions to namespace, e.g. type for T itself'''
        assert False  #pragma NO COVER
        return TypeScriptSourceCode('')
    pass

class ClassCodecImpl:
    t:type
    attr_codecs:dict[str,Any]  # codec
    custom_codec:CustomClassCodec|None = None
    def __init__(self, t:type, attr_codecs:dict[str,Any]):
        self.t=t
        self.attr_codecs=attr_codecs
        if issubclass(t,CustomClassCodec):
            self.custom_codec=t
        pass
    def encode(self,x,_:None|Callable[[Any],JsonType]) -> JsonType:
        'encode {x} as a {self.t}'
        try:
            if type(x) is not self.t:
                xt=type(x)
                raise Exception(f'{x!r} (of type {xt}) is not a {self.t}')
            if self.custom_codec is not None:
                return self.custom_codec.xju_json_codec_encode(x)
            def back_ref(x:Any) -> JsonType:
                return self.encode(x,None)
            result={}
            for n, attr_codec in self.attr_codecs.items():
                try:
                    result[n]=attr_codec.encode(getattr(x,n),back_ref)
                except Exception:
                    raise in_context(f'encode attribute {n}') from None
                pass
            return result
        except Exception:
            raise in_function_context(ClassCodecImpl.encode,vars()) from None
        pass
    def decode(self,x,_:None|Callable[[JsonType],Any]) -> object:
        'deocde {x} as a {self.t}'
        try:
            if self.custom_codec is not None:
                result=self.custom_codec.xju_json_codec_decode(x)
                assert isinstance(result,self.t), (repr(result),self.t)
                return result
            def back_ref(x:JsonType) -> object:
                return self.decode(x,None)
            attr_values={}
            for n, attr_codec in self.attr_codecs.items():
                if n in x:
                    try:
                        value=attr_codec.decode(x[n],back_ref)
                        attr_values[n]=value
                        pass
                    except Exception:
                        raise in_context(f'decode attribute {n}') from None
                    pass
                pass
            try:
                return self.t(**attr_values)
            except Exception:
                raise in_context(f'init {self.t} with keyword arguments {attr_values}') from None
        except Exception:
            raise in_function_context(ClassCodecImpl.decode,vars()) from None
        pass
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        if self.custom_codec is not None:
            return self.custom_codec.xju_json_codec_get_json_schema(definitions)
        fqn=self.get_type_fqn()
        self_ref=f'#/definitions/{fqn}'
        if not fqn in definitions:
            definitions[fqn]={
                'description': self.t.__name__,
                'type': 'object',
                'properties': {
                    n: attr_codec.get_json_schema(definitions, self_ref)
                    for n, attr_codec in self.attr_codecs.items()
                }
            }
            pass
        return {
            '$ref': self_ref
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return self.get_type_fqn()
    def ensure_typescript_defs(self,namespace)->None:
        '''ensure {namespace} has a typescript definition for {self.get_type_fqn()}'''
        typescript_fqn=[TypeScriptUQN(_) for _ in self.get_type_fqn().split('.')]
        target_namespace=namespace.get_namespace_of(typescript_fqn)
        typescript_type_name=typescript_fqn[-1]
        tt=self.typescript_type(None)
        def type_back_ref()->TypeScriptSourceCode:
            return TypeScriptSourceCode(self.typescript_type(None))
        def isa_back_ref(expression:TypeScriptSourceCode)->TypeScriptSourceCode:
            return self.get_typescript_isa(expression,namespace,None)
        def asa_back_ref(expression:TypeScriptSourceCode)->TypeScriptSourceCode:
            return self.get_typescript_asa(expression,namespace,None)
        back_refs=TypeScriptBackRefs(type_back_ref,isa_back_ref,asa_back_ref)
        if typescript_type_name not in target_namespace.defs:
            target_namespace.defs[typescript_type_name]=TypeScriptSourceCode(
                f"type {typescript_type_name} = {{\n"+
                ''.join([f"    {attr_name}: {indent(4,TypeScriptSourceCode(attr_codec.typescript_type(back_refs)))};\n"
                         for attr_name, attr_codec in self.attr_codecs.items()])+
                f"}};")
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function asInstanceOf{typescript_type_name}(v: any): {tt}\n"
                f"{{\n"
                f"    try{{\n"
                f"        if (Array.isArray(v)) throw new Error(`${{v}} is an array`);\n"
                f"        if (v == null) throw new Error(`${{v}} is not an object it is null`);\n"
                f"        if (typeof v !== 'object') throw new Error(`${{v}} is not an object it is a ${{typeof v}}`);\n"
                f"        const attr_asa=function(name:string, asa:any):any{{\n"
                f"            try{{\n"
                f"                asa(v[name]);\n"
                f"            }}\n"
                f"            catch(e:any){{\n"
                f"                throw new Error(`attribute ${{name}} is invalid because ${{e}}`);\n"
                f"            }}\n"
                f"        }}\n" +
                ''.join([f"        attr_asa('{attr_name}',(x:any)=>{indent(8,attr_codec.get_typescript_asa('x',namespace,back_refs))});\n"
                         for attr_name, attr_codec in self.attr_codecs.items()])+
                f"        return v as {tt};\n"
                f"    }}\n"
                f"    catch(e:any){{\n"
                f"        throw new Error(`${{v}} is not a {tt} because ${{e}}`);\n"
                f"    }}\n"
                f"}}")
            target_namespace.defs[TypeScriptUQN(f"isInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function isInstanceOf{typescript_type_name}(v:any): v is {tt}\n"
                f"{{\n"
                f"    return (\n"
                f"        !Array.isArray(v) &&\n"
                f"        v !== null &&"+
                f"        typeof v === 'object'"+
                ''.join([" &&\n        "+indent(8,attr_codec.get_typescript_isa(f'v["{attr_name}"]',namespace,back_refs))
                         for attr_name, attr_codec in self.attr_codecs.items()])+")"
                f"}}")
            pass
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        if self.custom_codec is not None:
            return self.custom_codec.xju_json_codec_get_typescript_isa(expression,namespace)
        self.ensure_typescript_defs(namespace)
        fqn=self.get_type_fqn().split('.')
        return TypeScriptSourceCode('.'.join(fqn[0:-1]+[f"isInstanceOf{fqn[-1]}({expression})"]))

    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        if self.custom_codec is not None:
            return self.custom_codec.xju_json_codec_get_typescript_asa(expression,namespace)
        self.ensure_typescript_defs(namespace)
        fqn=self.get_type_fqn().split('.')
        fqn=self.get_type_fqn().split('.')
        return TypeScriptSourceCode('.'.join(fqn[0:-1]+[f"asInstanceOf{fqn[-1]}({expression})"]))
    pass

@dataclass
class EnumValueCodecImpl:
    t:EnumMeta
    value:Enum
    value_codec:CodecImplProto

    def encode(self,x,back_ref:None|Callable[[Any],JsonType]) -> JsonType:
        'encode {x!r} as a {self.t.__name__}.{self.value.name}'
        try:
            return self.value_codec.encode(x.value,back_ref)
        except Exception:
            raise in_function_context(EnumValueCodecImpl.encode,vars()) from None
        pass
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]) -> Enum:
        'deocde {x!r} as {self.t.__name__}.{self.value.name}'
        try:
            value=self.value_codec.decode(x,back_ref)
            if value != self.value.value:
                raise Exception(f'{value!r} is not {self.value.value!r}')
            return self.value
        except Exception:
            raise in_function_context(EnumValueCodecImpl.decode,vars()) from None
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        result = self.value_codec.get_json_schema(definitions,self_ref)
        result.update({
            'enum': [ self.value_codec.encode(self.value.value,None) ]
        })
        return result
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        return f'{self.get_type_fqn()}.{self.value.name}'
    def typescript_value(self) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(json.dumps(self.value.value))
    def ensure_typescript_defs(self,namespace)->None:
        # EnumCodecImpl supplies defs
        pass
    def get_typescript_isa(
            self,
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace,
            back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        return TypeScriptSourceCode(f'({expression} === {self.typescript_type(back_refs)})')
    def get_typescript_asa(
            self,
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace,
            back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        tt=self.typescript_type(back_refs)
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{\n"
            f"    if (v !== {tt}) throw new Error(`the ${{typeof v}} ${{v}} is not {tt}`);\n"
            f"    return v as {tt};\n"
            f"}})({expression})")
    pass
    
class EnumCodecImpl:
    def __init__(self,t:EnumMeta, value_codecs:dict[str, EnumValueCodecImpl]):
        self.t=t
        self.value_codecs=value_codecs
        pass
    def encode(self,x:Enum,back_ref:None|Callable[[Any],JsonType]) -> JsonType:
        'encode {self.t} value {x!r}'
        try:
            if not isinstance(x, self.t):
                raise Exception(f'{x} (of type {x.__class__.__name__}) is not a {self.t.__name__}')
            return self.value_codecs[x.name].encode(x,back_ref)
        except Exception:
            raise in_function_context(EnumCodecImpl.encode,vars()) from None
        pass
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]) -> Enum:
        '''decode {x!r} as enum {self.t.__name__} value'''
        try:
            exceptions=list[BaseException]()
            for n, c in self.value_codecs.items():
                try:
                    return c.decode(x,back_ref)
                except Exception:
                    exceptions.append(in_context(f'decode as {self.t.__name__}.{n}'))
                    pass
                pass
            raise AllFailed(exceptions)
        except Exception:
            raise in_function_context(EnumCodecImpl.decode,vars()) from None
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        return {
            "oneOf": [
                codec.get_json_schema(definitions, self_ref) for codec in self.value_codecs.values() ]
        }
    def get_type_fqn(self):
        '''get the fully qualified name of {self.t}'''
        return get_type_fqn(self.t)
    def typescript_type(self,back_refs:TypeScriptBackRefs|None)->str:
        return self.get_type_fqn()
    def ensure_typescript_defs(self, namespace) -> None:
        typescript_fqn=[TypeScriptUQN(_) for _ in self.get_type_fqn().split('.')]
        target_namespace=namespace.get_namespace_of(typescript_fqn)
        typescript_type_name=typescript_fqn[-1]
        if typescript_type_name not in target_namespace.defs:
            enum_value_codec:EnumValueCodecImpl
            target_namespace.defs[typescript_type_name]=TypeScriptSourceCode(
                f"enum {typescript_type_name} {{\n"+
                ',\n'.join([f'    {name} = {enum_value_codec.typescript_value().value()}'
                         for name, enum_value_codec in self.value_codecs.items()])+
                f'\n}};\n')
            target_namespace.defs[TypeScriptUQN(f"asInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function asInstanceOf{typescript_type_name}(v: any): {typescript_type_name}\n"
                f"{{\n"
                f"    return {indent(4,self.get_typescript_asa(TypeScriptSourceCode('v'),namespace,None))};\n"
                f"}}")
            target_namespace.defs[TypeScriptUQN(f"isInstanceOf{typescript_type_name}")]=TypeScriptSourceCode(
                f"function isInstanceOf{typescript_type_name}(v:any): v is {typescript_type_name}\n"
                f"{{\n"
                f"    return {indent(4,self.get_typescript_isa(TypeScriptSourceCode('v'),namespace,None))};\n"
                f"}}")
        pass
    def get_typescript_isa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        enum_value_codec:EnumValueCodecImpl
        isas=[' ||\n       '+indent(12,enum_value_codec.get_typescript_isa(TypeScriptSourceCode(f"v"),
                                                                           namespace,back_refs))
              for enum_value_codec in self.value_codecs.values()]
        return TypeScriptSourceCode(
            f"((v:any): v is {self.typescript_type(back_refs)}=>{{\n"
            f"    return false" +
            f"".join(isas) + ";\n"+
            f"}})({expression})")
    def get_typescript_asa(self,
                           expression:TypeScriptSourceCode,
                           namespace: TypeScriptNamespace,
                           back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        enum_value_codec:EnumValueCodecImpl
        tt=self.typescript_type(back_refs)
        asas=[f"    try{{\n"+
              f"        {indent(8,enum_value_codec.get_typescript_asa(TypeScriptSourceCode('v'),namespace,back_refs))};\n"+
              f"        return v as {tt};\n"+
              f"    }}\n"+
              f"    catch(e:any){{\n"+
              f"        es.push(e.message);\n"+
              f"    }};\n"
              for enum_value_codec in self.value_codecs.values()]
        return TypeScriptSourceCode(
            f"((v: any): {tt} => {{try{{\n" +
            f"    var es = new Array<string>();\n"+
            "".join(asas)+
            f"    throw new Error(es.join(' and '));\n"+
            f"}}catch(e:any)\n"+
            f"{{\n"+
            f"    throw new Error(`${{v}} is not a {tt} because ${{e}}`);\n"+
            f"}}}})({expression})")
    pass

class SelfCodecImpl:
    def encode(self,x,back_ref:None|Callable[[Any],JsonType]) -> JsonType:
        'encode {x} as a Self'
        try:
            assert back_ref is not None
            return back_ref(x)
        except Exception:
            raise in_function_context(SelfCodecImpl.encode,vars()) from None
        pass
    def decode(self,x,back_ref:None|Callable[[JsonType],Any]) -> object:
        'deocde {x} as a Self'
        try:
            assert back_ref is not None
            return back_ref(x)
        except Exception:
            raise in_function_context(SelfCodecImpl.decode,vars()) from None
        pass
    def get_json_schema(self, definitions:dict[str,dict], self_ref:None|str) -> dict:
        assert self_ref is not None
        return {
            '$ref': self_ref
        }
    def typescript_type(self,back_refs:TypeScriptBackRefs|None) -> str:
        assert back_refs is not None
        return back_refs.type_back_ref().value()
    def ensure_typescript_defs(self,namespace)->None:
        # self would already have been done
        pass
    def get_typescript_isa(
            self,
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace,
            back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        assert back_refs is not None
        return back_refs.isa_back_ref(expression)
    def get_typescript_asa(
            self,
            expression:TypeScriptSourceCode,
            namespace: TypeScriptNamespace,
            back_refs:TypeScriptBackRefs|None) -> TypeScriptSourceCode:
        assert back_refs is not None
        return back_refs.asa_back_ref(expression)
    pass
    
def _explodeSchema(t:type|NewType|TypeVar|GenericAlias|UnionType|_LiteralGenericAlias|_GenericAlias|Enum|Type[Enum],
                   type_var_map:dict[TypeVar,Any]|None):
    '''explode type {t!r} into a tree of codecs using map {type_var_map} to resolve any generic type refs i.e. TypeVars'''
    try:
        if type(t) is TypeVar:
            assert type_var_map is not None
            assert t in type_var_map, type_var_map.keys()
            return _explodeSchema(type_var_map[t],type_var_map)
        if t is float or (isinstance(t,NewType) and t.__supertype__ is float):
            return NoopCodecImpl[float](float)
        if t is int or (isinstance(t,NewType) and t.__supertype__ is int):
            return NoopCodecImpl[int](int)
        if t is str or (isinstance(t,NewType) and t.__supertype__ is str):
            return NoopCodecImpl[str](str)
        if t is bytes or (isinstance(t,NewType) and t.__supertype__ is bytes):
            return BytesCodecImpl(type_var_map)
        if t is bool or (isinstance(t,NewType) and t.__supertype__ is bool):
            return NoopCodecImpl[bool](bool)
        if t is None or t is NoneType:
            return NoneCodecImpl()
        if t is list:
            return AnyListCodecImpl()
        if type(t) is GenericAlias and get_origin(t) is list:
            return ListCodecImpl(_explodeSchema(get_args(t)[0],type_var_map))
        if type(t) is _LiteralGenericAlias and len(t.__args__)==1:
            return _explode_literal(t.__args__[0])
        if type(t) is _LiteralGenericAlias:
            return UnionCodecImpl(get_args(t),
                              {t:_explode_literal(t) for t in get_args(t)})
        if type(t) is GenericAlias and get_origin(t) is tuple:
            return TupleCodec([_explodeSchema(_,type_var_map) for _ in get_args(t)])
        if t is dict:
            return AnyDictCodecImpl()
        if type(t) is GenericAlias and get_origin(t) is dict:
            return DictCodecImpl(*[_explodeSchema(_, type_var_map) for _ in get_args(t)])
        if type(t) is UnionType:
            return UnionCodecImpl(get_args(t),
                              {t:_explodeSchema(t, type_var_map) for t in get_args(t)})
        if type(t) is _UnionGenericAlias:
            return UnionCodecImpl(get_args(t),
                              {t:_explodeSchema(t, type_var_map) for t in get_args(t)})
        if t is Self:
            return SelfCodecImpl()
        if type(t) is _GenericAlias:
            local_type_var_map={
                type_var: value
                for type_var, value in
                list((type_var_map or {}).items())+list(zip(get_origin(t).__parameters__, get_args(t)))
            }
            return ClassCodecImpl(
                get_origin(t),
                { n: _explodeSchema(nt,local_type_var_map)
                  for n,nt in get_type_hints(get_origin(t)).items()})
        if type(t) is EnumType and issubclass(t, Enum):
            return EnumCodecImpl(
                t,{name: _explodeSchema(vv,type_var_map)
                   for name,vv in t.__members__.items()})
        if isinstance(t,Enum):
            return EnumValueCodecImpl(t.__class__,t,_explodeSchema(type(t.value),type_var_map))
        assert isinstance(t,type), (type(t), t)
        if issubclass(t,xju.newtype.Int):
            return NewIntCodecImpl(t)
        if issubclass(t,xju.newtype.Float):
            return NewFloatCodecImpl(t)
        if issubclass(t,xju.newtype.Str):
            return NewStrCodecImpl(t)
        return ClassCodecImpl(
            t,{n: _explodeSchema(nt,type_var_map) for n,nt in get_type_hints(t).items()})
    except Exception:
        raise in_function_context(_explodeSchema,vars()) from None
    pass

def _explode_literal(value: Any):
    '''create codec for literal value {value!r}'''
    try:
        if type(value) is str:
            return LiteralStrCodecImpl(value)
        if type(value) is bool:
            return LiteralBoolCodecImpl(value)
        if type(value) is int:
            return LiteralIntCodecImpl(value)
        t=type(value)
        raise Exception(f'{t} literals are not supported (only support str, int, bool)')
    except:
        raise in_function_context(_explode_literal, vars())
    pass

def indent(n: int, s:TypeScriptSourceCode)->str:
    lines=s.splitlines()
    return '\n'.join([lines[0].value()]+[(' '*n)+l.value() for l in lines[1:]])

def get_type_fqn(t:type) -> str:
    if t.__module__=='__main__':
        return t.__name__
    return f"{t.__module__}.{t.__name__}"

# for development of mypy plugin, to see what return type should be for a case
# see json_codec_mypy_plugin.show_return_type
# class _xxx:
#     pass



from typing import Callable
from mypy.plugin import (
    Plugin, FunctionSigContext, FunctionContext, Expression,
    CheckerPluginInterface
)
from mypy.nodes import (
    Expression, NameExpr, IndexExpr, TupleExpr, StrExpr, IntExpr, FloatExpr, OpExpr, MemberExpr,
    TypeInfo,
    TypeAlias
)
from mypy.types import (
    AnyType,
    FunctionLike,
    Instance,
    LiteralType,
    NoneType,
    TupleType,
    Type,
    TypeAliasType,
    TypeOfAny,
    UninhabitedType,
    UnionType,
)
from mypy.typeops import tuple_fallback
from mypy.checker import TypeChecker

CODEC_FQN="xju.json_codec.codec"


def adjust_codec_signature(s: FunctionSigContext) -> FunctionLike:
    """adjust signature to Any -> Any so we can param type after mypy resolves param type"""
    if len(s.args) != 1:
        l=len(s.args)
        s.api.fail(f"{CODEC_FQN}() takes exactly one parameter, not {l}", s.context)
        return s.default_signature
    s.default_signature.arg_types[0] = AnyType(TypeOfAny.explicit)
    return s.default_signature

def flatten_union_terms(expr: OpExpr) -> list[Expression] | UninhabitedType:
    if expr.op != "|":
        return UninhabitedType()
    left = flatten_union_terms(expr.left) if isinstance(expr.left, OpExpr) else [expr.left]
    right = flatten_union_terms(expr.right) if isinstance(expr.right, OpExpr) else [expr.right]
    if isinstance(left, UninhabitedType) or isinstance(right, UninhabitedType):
        return UninhabitedType()
    return left+right

def infer_literal_type(expr: Expression, checker_api:CheckerPluginInterface) -> LiteralType:
    try:
        match expr:
            case StrExpr():
                return LiteralType(
                    expr.value,
                    builtin_type(checker_api,'str'))
            case IntExpr():
                return LiteralType(
                    expr.value,
                    builtin_type(checker_api,'int'))
            case FloatExpr():
                return LiteralType(
                    expr.value,
                    builtin_type(checker_api,'float'))
            # surprise... bool literal e.g. True is represented as a
            # NameExpr with name "True"
            case NameExpr():
                return LiteralType(
                    expr.name == "True",
                    builtin_type(checker_api,'bool'))
        raise Exception(f"{expr} is not one of StrExpr, IntExpr, FloatExpr, NameExpr")
    except Exception as e:
        raise Exception(f"failed to infer type for presumed literal {expr} becase {e}") from None

def infer_codec_value_type(arg_expr: Expression | list[Expression],
                           checker_api:CheckerPluginInterface) -> Type:
    #import pdb; pdb.set_trace()
    try:
        match arg_expr:
            case list():
                if len(arg_expr) != 1:
                    return UninhabitedType()
                return infer_codec_value_type(arg_expr[0], checker_api)
            case MemberExpr():
                return infer_codec_value_type(arg_expr.expr, checker_api)
            case NameExpr() if arg_expr.fullname == 'types.NoneType':
                return NoneType()
            case NameExpr() if arg_expr.name == 'None':
                # REVISIT: tell them to use NoneType
                return UninhabitedType()
            case NameExpr():
                # general named type or alias
                return named_type(checker_api,arg_expr)
            #?case StrExpr():
            case IndexExpr():
                if not isinstance(arg_expr.base, NameExpr):
                    return UninhabitedType()
                if arg_expr.base.name=="Literal" and not isinstance(arg_expr.index, TupleExpr):
                    # single literal value
                    return infer_literal_type(arg_expr.index, checker_api)
                if arg_expr.base.name=="Literal":
                        assert isinstance(arg_expr.index, TupleExpr)
                        # multi-value literal
                        return UnionType([
                            infer_literal_type(a, checker_api) for a in arg_expr.index.items])
                if isinstance(arg_expr.index, TupleExpr):
                    if arg_expr.base.name=="tuple":
                        items=[infer_codec_value_type(a, checker_api) for a in arg_expr.index.items]
                        return TupleType(
                            items,
                            builtin_type(checker_api,"tuple"))
                    # other e.g. dict[str,int]
                    return checker_api.named_generic_type(
                        arg_expr.base.name,
                        [infer_codec_value_type(a, checker_api) for a in arg_expr.index.items]
                    )
                # only one type param, e.g. list[int]
                return checker_api.named_generic_type(
                    arg_expr.base.name,
                    [infer_codec_value_type(arg_expr.index, checker_api)]
                )
            case OpExpr() if arg_expr.op == "|":
                # union
                terms = flatten_union_terms(arg_expr)
                if isinstance(terms, UninhabitedType):
                    return terms
                return UnionType([infer_codec_value_type(a, checker_api) for a in terms])
        return UninhabitedType()
    except Exception as e:
        raise Exception(f"failed to infer xju.json_codec.codec value type from expression {arg_expr} because {e}") from None

def adjust_codec_return_type(x: FunctionContext) -> Type:
    arg_type=x.arg_types[0]
    arg_expr=x.args[0]
    inferred_codec_type = infer_codec_value_type(arg_expr, x.api)
    if isinstance(inferred_codec_type, UninhabitedType):
        arg_value_str=[str(_) for _ in arg_expr]
        return_type=x.default_return_type
        typeof_return_type=type(return_type)
        x.api.fail(f"type {arg_type} is not supported by {CODEC_FQN}(); note the argument has value expression {arg_value_str}; note the return type is {typeof_return_type}({return_type})", x.context)
        return AnyType(TypeOfAny.from_error)
    assert isinstance(x.default_return_type, Instance)
    x.default_return_type.args=(inferred_codec_type, )
    return x.default_return_type

# for development to figure out what return type should be in each case via pdb
def show_return_type(x: FunctionContext) -> Type:
    assert isinstance(x.default_return_type, Instance)
    return_type=x.default_return_type
    typeof_return_type=type(return_type)
    #import pdb; pdb.set_trace()
    return return_type

class JsonCodecPlugin(Plugin):
    def get_function_hook(self, fullname: str) -> Callable[[FunctionContext], Type] | None:
        if fullname==CODEC_FQN:
            return adjust_codec_return_type
        if fullname=="xju.json_codec._xxx":
            return show_return_type
        return None
    
    def get_function_signature_hook(
            self, fullname: str
    ) -> Callable[[FunctionSigContext], FunctionLike] | None:
        if fullname==CODEC_FQN:
            return adjust_codec_signature
        return None

def plugin(version: str):
    # ignore version argument if the plugin works with all mypy versions.
    return JsonCodecPlugin


def builtin_type(checker_api:CheckerPluginInterface, name: str) -> Instance:
    """lookup built-in type {name}"""
    try:
        # need more info than CheckerPluginInterface exposes
        assert isinstance(checker_api, TypeChecker), (type(checker_api), checker_api)
        # checker_api.named_type is good enough for builtins
        return checker_api.named_type(name)
    except Exception as e:
        raise Exception(f"failed to lookup built-in type {name} becase {e}")


def named_type(checker_api:CheckerPluginInterface, expr: NameExpr) -> Instance | TypeAliasType:
    """lookup type of expression {expr}
    
    - name can be an alias -> TypeAliasType
    - otherwise -> Instance
    """
    try:
        # need more info than CheckerPluginInterface exposes
        assert isinstance(checker_api, TypeChecker), (type(checker_api), checker_api)

        # checker_api.named_type is not good enough as it does not handle aliases properly

        # apparently checker_api keeps context, so we have to give the unqualified name
        # not the fullname
        node = checker_api.lookup_qualified(expr.name).node
        if isinstance(node, TypeAlias):
            return TypeAliasType(node, [])
        assert isinstance(node, TypeInfo), f"{expr} is not a type, it is the {node} {node.__class__}"
        return Instance(node, [])
    except Exception as e:
        raise Exception(f"failed to lookup type named by {expr} becase {e}") from None




# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
from typing import Union,List,Dict
import json
from xju.newtype import Int

class ByteCountTag:pass
class ByteCount(Int[ByteCountTag]):
    pass

def toJson(x:Union[str,int,float,bool,None,List,Dict],cls=json.JSONEncoder) -> str:
    '''JSON-encode {x}
       - produces multi-line format with 4-space indent'''
    return json.dumps(x,sort_keys=True,indent=4,separators=(',',': '),cls=cls)

def fromJson(x:str) -> Union[str,int,float,bool,None,List,Dict]:
    '''JSON-decode {x}'''
    return json.loads(x)



# coding: utf-8
#
# Copyright (c) 2018 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# jquery-like python library
#
# parse a HTML text into a tree, with search, manipulation
# and re-output as HTML text
#
# parse() and parseFile() are the primary interface.
#
# REVISIT: need better diags, eg:
#  track selectors used to find/filter nodes and report them
#  on failure eg replace/html
#
from __future__ import annotations
from html.parser import HTMLParser
from html.entities import entitydefs as htmlentitydefs
from html.entities import name2codepoint as entities
from html.entities import codepoint2name as reverseentities
import sys
import traceback
import string
from xju.xn import in_context,in_function_context

class Pos:
    def __init__(self, file, line, col):
        self.file=file
        self.line=line
        self.col=col
        return
    def __str__(self):
        return '%(file)s:%(line)s:%(col)s' % self.__dict__
    pass

class ParseFailed(Exception):
    def __init__(self, cause, pos):
        self.cause=cause
        self.pos=pos
        Exception.__init__(self,self.__str__())
        return
    def __str__(self):
        return 'failed to parse html at %(pos)s because\n%(cause)s'%self.__dict__
    pass


def encodeEntity(c):
    x=reverseentities.get(ord(c),None)
    if x is None: return c
    return u'&%(x)s;'%vars()

def encodeEntities(s:str)->str:
    if s is None: return u''
    x=u''.join([encodeEntity(_) for _ in s])
    return x

class Node:
    def __init__(self, parent=None):
        self.parent=parent
        if parent:
            parent.children.append(self)
            pass
        pass
    def predecessor(self):
        if self.parent:
            i=self.parent.children.index(self)
            if i>0:
                return [self.parent.children[i-1]]
            pass
        return []
    def successor(self):
        if self.parent:
            i=self.parent.children.index(self)+1
            if i<len(self.parent.children):
                return [self.parent.children[i]]
            pass
        return []
    
class Root(Node):
    def __init__(self):
        Node.__init__(self)
        self.children:list[Node]=[]
    def indexOf(self, child):
        return [ _[0] for _ in zip(range(0,len(self.children)),
                                   self.children) if _[1]==child][0]
    def remove(self, node):
        i=self.indexOf(node)
        self.children=self.children[0:i]+self.children[i+1:]
        node.parent=None
        return self
    pass

endOptional=frozenset([
        'body',
        'colgroup',
        'dd',
        'dt',
        'head',
        'html',
        'li',
        'option',
        'p',
        'tbody',
        'td',
        'tfoot',
        'th',
        'thead',
        'tr'])

emptyTags=frozenset([
        'area',
        'base',
        'basefont',
        'br',
        'col',
        'frame',
        'hr',
        'img',
        'input',
        'isindex',
        'link',
        'meta',
        'param'])

class Tag(Node):
    def __init__(self, tagName, attrs, parent, pos):
        Node.__init__(self, parent)
        self.pos=pos
        self.tagName=tagName
        self.attrs=dict(attrs)
        self.children:list[Node]=[]
        self.end=''
        self.classes=set(self.attrs.get('class','').split())
        return
    def __repr__(self):
        return '%(tagName)s at %(pos)s' % self.__dict__
    def __str__(self):
        encodedAttrs=['%s="%s"' % (_[0],encodeEntities(_[1])) for 
                      _ in self.attrs.items()]
        encodedAttrs.sort()
        start=' '.join([self.tagName]+encodedAttrs)
        if self.tagName in ['script','style']:
            content=self.text()
        else:
            content=''.join([str(_) for _ in self.children])
            pass
        end=self.end
        return '''<%(start)s>%(content)s%(end)s'''%vars()
    def hasClass(self,c):
        return c in self.classes
    def addClass(self,c):
        self.classes.add(c)
        self.attrs['class']=' '.join(self.classes)
        return self
    def removeClass(self,c):
        if c in self.classes:
            self.classes.remove(c)
            pass
        self.attrs['class']=' '.join(self.classes)
        if len(self.classes)==0:
            del self.attrs['class']
        return self
    def attr(self, a, val=None):
        if not val is None:
            self.attrs[a]=str(val)
        return self.attrs.get(a,'')
    def removeAttr(self, a):
        if a in self.attrs: del self.attrs[a]
        return
    def attrEquals(self, a, val):
        return self.attrs.get(a,None)==val
    def hasAttr(self, a):
        return a in self.attrs
    def indexOf(self, child):
        return [ _[0] for _ in zip(range(0,len(self.children)),
                                   self.children) if _[1]==child][0]
    def remove(self, node):
        i=self.indexOf(node)
        self.children=self.children[0:i]+self.children[i+1:]
        node.parent=None
        return self
    def replace(self, node, newNode):
        if newNode.parent:
            newNode.parent.remove(newNode)
        self.children[self.indexOf(node)]=newNode
        newNode.parent=self
        return self
    def clone(self, newParent):
        result=Tag(self.tagName, self.attrs.items(), newParent, self.pos)
        result.end=self.end
        result.classes=set(self.classes)
        for c in self.children:
            assert isinstance(c,Tag) or \
                isinstance(c,Data) or \
                isinstance(c,Comment) or \
                isinstance(c,Decl) or \
                isinstance(c,PI)
            c.clone(result)
        return result
    def text(self):
        if self.tagName=='br':
            return u'\n'
        if self.tagName=='p':
            return u''.join([_.text() for _ in self.children])+u'\n'
        if self.tagName=='li':
            return u''.join([_.text() for _ in self.children])+u'\n'
        return u''.join([_.text() for _ in self.children])
    pass

class Data(Node):
    '''CData, including <script> and <style> content'''
    def __init__(self, data, parent, pos):
        Node.__init__(self, parent)
        self.pos=pos
        self.data=data
        pass
    def __str__(self):
        return encodeEntities(self.data)
    def __repr__(self):
        return 'data at %(pos)s, %(data)r' % self.__dict__
    def clone(self, newParent):
        result=Data(self.data, newParent, self.pos)
        return result
    def text(self):
        return self.data
    pass

class Comment(Node):
    '''Comment'''
    def __init__(self, comment, parent, pos):
        Node.__init__(self, parent)
        self.pos=pos
        self.comment=comment
    def __str__(self):
        return '<!--%(comment)s-->' % self.__dict__
    def __repr__(self):
        return 'comment at %(pos)s' % self.__dict__
    def clone(self, newParent):
        result=Comment(self.comment, newParent, self.pos)
        return result
    def text(self):
        return u''
    pass

class Decl(Node):
    '''Decl'''
    def __init__(self, decl, parent, pos):
        Node.__init__(self, parent)
        self.pos=pos
        self.decl=decl
    def __str__(self):
        return '<!%(decl)s>' % self.__dict__
    def __repr__(self):
        return 'decl at %(pos)s' % self.__dict__
    def clone(self, newParent):
        result=Decl(self.decl, newParent, self.pos)
        return result
    def text(self):
        return u''
    pass

class PI(Node):
    '''Processing Instruction'''
    def __init__(self, pi, parent, pos):
        Node.__init__(self, parent)
        self.pos=pos
        self.pi=pi
    def __str__(self):
        return '<?%(pi)s>' % self.__dict__
    def __repr__(self):
        return 'processing instruction at %(pos)s' % self.__dict__
    def clone(self, newParent):
        result=PI(self.pi, newParent, self.pos)
        return result
    def text(self):
        return u''
    pass

class Parser(HTMLParser):
    def __init__(self, fileName):
        HTMLParser.__init__(self)
        self.fileName=fileName
        self.root=Root()
        self.current:Root|Tag=self.root
        return
    def pos(self):
        return Pos(self.fileName, *self.getpos())
    def handle_starttag(self, tag, attrs):
        if tag in emptyTags:
            Tag(tag, attrs, self.current, self.pos())
        else:
            #end-optional tag ends current tag of same name, eg
            #<li>x
            #<li>y
            if isinstance(self.current, Tag) and \
                    tag in endOptional and \
                    self.current.tagName == tag:
                self.current=self.current.parent
            self.current=Tag(tag, attrs, self.current, self.pos())
        return
    def handle_endtag(self, tag):
        current=self.current
        while not isinstance(current,Root) and current.tagName != tag:
            current=current.parent
            pass
        if not isinstance(current,Root):
            self.current=current
            self.current.end=u'</%(tag)s>'%vars()
            self.current=self.current.parent
        return
    def handle_data(self,data):
        Data(data, self.current, self.pos())
        return
    def handle_comment(self,comment):
        Comment(comment, self.current, self.pos())
        return
    def handle_decl(self,decl):
        Decl(decl,self.current, self.pos())
        return
    def handle_pi(self,pi):
        PI(pi,self.current, self.pos())
        return

def filter(node, predicate):
    '''return [node if matches predicate] + all children that match'''
    result=[_ for _ in [node] if predicate(_)]
    if isinstance(node,Tag):
        for c in node.children:
            result.extend(filter(c, predicate))
    return result
        
class Selection:
    nodeList:list[Node]
    def __init__(self, nodeList:Selection|Node|list[Node]):
        if isinstance(nodeList,Selection):
            self.nodeList=nodeList.nodeList
        elif isinstance(nodeList,Node):
            self.nodeList=[nodeList,]
        else:
            self.nodeList=nodeList[:]
            pass
        return
    def find(self, predicate):
        '''find all children of our nodes that match predicate'''
        result=[]
        for _ in self.nodeList:
            result.extend(filter(_,predicate))
        return Selection(result)
    def filter(self, predicate):
        '''nodes of ours that match predicate'''
        return Selection([_ for _ in self.nodeList if predicate(_)])
    def unless(self, predicate):
        '''nodes of ours that don't match predicate'''
        return Selection([_ for _ in self.nodeList if not predicate(_)])
    def html(self, nodes):
        '''replace our first node's children with the specified list of nodes/html string'''
        if type(nodes)==str:
            nodes=parse(nodes)
            pass
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
            pass
        for n in nodes:
            if n.parent: n.parent.remove(n)
        for n in self.nodeList[0:1]:
            n.children=nodes
            for c in n.children:
                c.parent=n
        return self
    def text(self, s=None):
        '''replace our first node's children with the specified text string'''
        '''or return concatenation of text content of children'''
        if s is None:
            return u''.join([_.text() for _ in self.nodeList])
        ss=parse(encodeEntities(str(s)))
        for n in self.nodeList:
            Selection([n]).html(ss)
        return self
    def replace(self, nodes):
        '''replace first of specified nodes with first of our nodes'''
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
        nodes[0].parent.replace(nodes[0],self.nodeList[0])
        return self
    def appendTo(self, nodes):
        '''append our nodes to children of first of specified nodes'''
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
        for n in self.nodeList:
            n.parent=nodes[0]
            nodes[0].children.append(n)
        return self
    def addAfter(self, nodes):
        '''add our nodes after first of specified nodes'''
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
        parent=nodes[0].parent
        index=parent.indexOf(nodes[0])
        for n in self.nodeList:
            n.parent=parent
        parent.children[index+1:index+1]=self.nodeList
        return self
    def addBefore(self, nodes):
        '''add our nodes before first of specified nodes'''
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
        parent=nodes[0].parent
        index=parent.indexOf(nodes[0])
        for n in self.nodeList:
            n.parent=parent
        parent.children[index:index]=self.nodeList
        return self
    def detach(self):
        '''remove each of our nodes from its parent'''
        for n in self.nodeList:
            if n.parent:
                n.parent.remove(n)
        return self
    def remove(self):
        return self.detach()
    def empty(self):
        '''remove all children from each of our nodes'''
        for n in self.nodeList:
            if isinstance(n,Tag):
                for c in n.children:
                    c.parent=None
                    pass
                n.children=[]
                pass
            pass
        return self
    def first(self):
        '''return Selection containing first of our nodes'''
        return Selection(self.nodeList[0:1])
    def last(self):
        '''return Selection containing first of our nodes'''
        return Selection(self.nodeList[-1:])
    def children(self):
        '''return Selection containing children of our nodes'''
        nodeList:list[Node]=[]
        return Selection(sum([_.children for _ in self.nodeList],nodeList))
    def predecessors(self):
        '''return Selection containing predecessor of each of our nodes'''
        nodeList:list[Node]=[]
        return Selection(sum([_.predecessor() for _ in self.nodeList],nodeList))
    def successors(self):
        '''return Selection containing successor of each of our nodes'''
        nodeList:list[Node]=[]
        return Selection(sum([_.successor() for _ in self.nodeList],nodeList))
    def clone(self):
        '''return Selection containing a copy of our nodes'''
        return Selection([_.clone(None) for _ in self.nodeList])
    def addClass(self, name):
        '''add class %(name)s to each of our children'''
        for n in self.nodeList:
            if isinstance(n,Tag):
                n.addClass(name)
                pass
            pass
        return self
    def removeClass(self, name):
        '''remove class %(name)s from each of our children'''
        for n in self.nodeList:
            if isinstance(n,Tag):
                n.removeClass(name)
                pass
            pass
        return self
    def hasClass(self,c):
        '''True iff all our nodes have class c'''
        each=[True for _ in self.nodeList if isinstance(_,Tag) and _.hasClass(c)]
        return len(each)==len(self.nodeList)
    def attr(self, name, value=None, joiner=u''):
        '''attr('src') gets the values of the src attributes of our nodes and joins them with joiner, returning a single string'''
        """attr('src','fred') sets the src attribute of our only node to 'fred'"""
        if value is None:
            return joiner.join([_.attr(name, value) for _ in self.nodeList if isinstance(_,Tag)])
        [_.attr(name, value) for _ in self.nodeList if isinstance(_,Tag)]
        return self
    def attrs(self, name, value=None):
        '''attrs('src') lists the values of the src attributes of each of our nodes'''
        """attrs('src','fred.html') sets the src attribute of each of our nodes to 'html'"""
        if value is None:
            return [_.attr(name, value) for _ in self.nodeList if isinstance(_,Tag)]
        [_.attr(name, value) for _ in self.nodeList if isinstance(_,Tag)]
        return self
    def removeAttr(self, name):
        [_.removeAttr(name) for _ in self.nodeList if isinstance(_,Tag)]
        return self
    def join(self,nodes):
        '''return new Selection containing nodes with each separated by a clone of our nodes'''
        if isinstance(nodes, Selection):
            nodes=nodes.nodeList
            pass
        resultNodes=[]
        for i,n in enumerate(nodes):
            resultNodes.append(n)
            if i+1 < len(nodes): resultNodes.extend(self.clone().nodeList)
            pass
        return Selection(resultNodes)
    def __str__(self):
        '''re-format as html'''
        return ''.join([str(_) for _ in self.nodeList])
    def __len__(self):
        return len(self.nodeList)
    def __getitem__(self, key):
        return Selection(self.nodeList.__getitem__(key))
    def __add__(self, b):
        if isinstance(b,Selection):
            return Selection(self.nodeList+b.nodeList)
        return NotImplemented
    def utf8(self):
        return str(self).encode('utf-8')
    pass

# basic predicates
def hasClass(c:str):
    return lambda node: isinstance(node, Tag) and node.hasClass(c)
def tagName(t:str):
    return lambda node: isinstance(node, Tag) and node.tagName==t
def attrEquals(attr:str,value:str):
    return lambda node: isinstance(node, Tag) and node.attrEquals(attr,value)
def hasAttr(attr):
    return lambda node: isinstance(node, Tag) and node.hasAttr(attr)

def parse(s:str, origin='unknown') -> Selection:
    '''parse HTML string "%(origin)s"'''
    parser=Parser(origin)
    try:
        u=str(s)
        parser.feed(u)
        parser.close()
        result=Selection(parser.root.children)
        result.detach()
        return result
    except:
        raise ParseFailed(str(''.join(traceback.format_tb(sys.exc_info()[2])))+'\n'+str(sys.exc_info()[1]), parser.pos()) from None
    pass

def parseFile(fileName,encoding='utf-8') -> Selection:
    '''parse {fileName} as HTML'''
    try:
        with open(fileName,encoding=encoding) as f:
            return parse(f.read(),fileName)
        pass
    except Exception:
        raise in_function_context(parseFile,vars())
    pass

# deprecated, use parseFile
def loadFile(fileName,encoding='utf-8'):
    return parseFile(fileName,encoding)
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

from xju.assert_ import Assert

class PatchAttr:
    '''patch attribute {attr} of {x} for context duration with value {temporary_value}'''
    def __init__(self,x:object,attr:str,temporary_value):
        Assert(x).hasAttr(attr)
        self.x=x
        self.attr=attr
        self.temporary_value=temporary_value
        pass

    def __enter__(self):
        self.orig=getattr(self.x,self.attr)
        setattr(self.x,self.attr,self.temporary_value)
        return self

    def __exit__(self,t,e,b):
        setattr(self.x,self.attr,self.orig)
        pass
    pass
